"""Testing python apps
"""
from parsl import *
import time

# parsl.set_stream_logger()
localThreads = {
    "sites": [
        {"site": "Local_Threads",
         "auth": {"channel": None},
         "execution": {
             "executor": "threads",
             "provider": None,
             "maxThreads": 10
         }
         }],
}

dfk = DataFlowKernel(config=localThreads)


@App('python', dfk)
def sleep_python(x):
    import time
    time.sleep(x)
    return x


@App('bash', dfk)
def sleep_bash(x):
    return 'sleep {0}'


def test_parallel_sleep_bash(n=10, sleep_dur=2, tolerance=0.3):
    """Indirect test to ensure that 10 threads are live using bash apps
    """

    start = time.time()

    d = []
    for i in range(0, n):
        d.extend([sleep_bash(sleep_dur)])

    for i in d:
        i.result()
    end = time.time()
    delta = end - start
    print("Sleep time : {0}, expected ~{1}+/- 0.3s".format(delta, sleep_dur))
    assert delta > sleep_dur - tolerance, "Slept too little"
    assert delta < sleep_dur + tolerance, "Slept too much"


def test_parallel_sleep_python(n=10, sleep_dur=2, tolerance=0.3):
    """Indirect test to ensure that 10 threads are live using python sleep apps
    This works only because the 10 threads are essentially sleeping and not
    doing manipulation of python objects and causing serialization via GIL.
    """

    start = time.time()

    d = []
    for i in range(0, n):
        d.extend([sleep_python(sleep_dur)])

    for i in d:
        i.result()
    end = time.time()
    delta = end - start
    print("Sleep time : {0}, expected ~{1}+/- 0.3s".format(delta, sleep_dur))
    assert delta > sleep_dur - tolerance, "Slept too little"
    assert delta < sleep_dur + tolerance, "Slept too much"


if __name__ == "__main__":

    test_parallel_sleep_bash()
    test_parallel_sleep_python()



import parsl
from parsl import *
import os

os.environ['CORI_USERNAME'] = 'yadunand'
from cori import multiNodeMPI as config
parsl.set_stream_logger()
dfk = DataFlowKernel(config=config)


@App("python", dfk)
def python_app_slow(duration):
    import platform
    import time
    time.sleep(duration)
    return "Hello from {0}".format(platform.uname())


def test_python_remote(count=10):
    """ Run with no delay
    """
    fus = []
    for i in range(0, count):
        fu = python_app_slow(0)
        fus.extend([fu])

    for fu in fus:
        print(fu.result())


def test_python_remote_slow(count=20):

    fus = []
    for i in range(0, count):
        fu = python_app_slow(count)
        fus.extend([fu])

    for fu in fus:
        print(fu.result())


@App("bash", dfk)
def bash_mpi_app(stdout=None, stderr=None):
    return """ls -thor
mpi_hello
    """


if __name__ == "__main__":

    items = []
    for i in range(0, 4):
        x = bash_mpi_app(stdout="parsl.{0}.out".format(i),
                         stderr="parsl.{0}.err".format(i))
        items.extend([x])

    for i in items:
        print(i.result())



# A "discriminator" is a point in a workflow process that waits for one of the incoming branches to complete before
# beginning the subsequent activity

from parsl import *
import random
import argparse

workers = ThreadPoolExecutor(max_workers=10)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    return random.randint(1, 10)


@App('python', dfk)
def square(x):
    return x**2


@App('python', dfk)
def cubed(x):
    return x**3


def test_discriminator(x=4):
    squares = []
    cubes = []
    total = []
    for i in range(x):
        squares.append(square(rand().result()))
        cubes.append(cubed(rand().result()))
        while squares[i].done() is False and cubes[i].done() is False:
            pass
        total.append(squares[i].result() + cubes[i].result())
    for i in range(x):
        print(total[i])


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", "--num1", default="4",
                        action="store", dest="x", type=int)
    args = parser.parse_args()
    test_discriminator(args.x)



from parsl import *
import argparse

workers = ThreadPoolExecutor(max_workers=5)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def fibonacci(n):
    if n == 0:
        return 0
    if n == 2 or n == 1:
        return 1
    else:
        return fibonacci(n - 1).result() + fibonacci(n - 2).result()


def test_fibonacci(x=5):
    results = []
    for i in range(x):
        results.append(fibonacci(i))
    for j in range(len(results)):
        print(results[j].result())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--num", default='5',
                        action="store", dest="a", type=int)
    args = parser.parse_args()
    test_fibonacci(args.a)



from parsl import *

workers = ThreadPoolExecutor(max_workers=4)


@App('bash', workers)
def echo(inputs=[], stderr='std.err', stdout='std.out'):
    cmd_line = 'echo {inputs[0]} {inputs[1]}'
    return cmd_line


@App('bash', workers)
def sleep_n(t):
    cmd_line = 'sleep {t}'
    return cmd_line


@App('bash', workers)
def cats_n_sleep(x, inputs, outputs):
    cmd_line = 'sleep $(($RANDOM % {x})); cat {inputs[0]} > {outputs[0]}'
    return cmd_line


@App('bash', workers)
def incr(inputs, outputs):
    cmd_line = 'y=$(cat {inputs[0]}); echo $(($y+1)) > {outputs[0]}'
    return cmd_line


@App('bash', workers)
def raise_error(inputs, outputs):
    cmd_line = 'y=$(cat {inputs[0]}); echo $(($y+1)) > {outputs[0]}'
    return cmd_line



"""Testing bash apps
"""
import parsl
from parsl import *

print("Parsl version: ", parsl.__version__)

import time
import argparse

workers = ThreadPoolExecutor(max_workers=10)
dfk = DataFlowKernel(executors=[workers])


@App('bash', dfk)
def sleep_foo(sleepdur, stdout=None):
    return """sleep {0}"""


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    fus = {}

    start = time.time()
    for i in range(int(args.count)):

        fus[i] = sleep_foo(5)

    print([(key, fus[key].result()) for key in fus])
    end = time.time()
    print("Total time : ", end - start)



from parsl import *
import parsl
import libsubmit

parsl.set_stream_logger()

print(parsl.__version__)
print(libsubmit.__version__)

from local import localIPPReuse as config
dfk = DataFlowKernel(config=config)


@App("python", dfk)
def python_app():
    import platform
    return "Hello from {0}".format(platform.uname())


@App("bash", dfk)
def bash_app(stdout=None, stderr=None):
    return 'echo "Hello from $(uname -a)" ; sleep 2'


def test_python():
    """Testing basic python functionality."""

    results = {}
    for i in range(0, 2):
        results[i] = python_app()

    print("Waiting ....")
    print(results[0].result())


def test_bash():
    """Testing basic bash functionality."""

    import os
    fname = os.path.basename(__file__)

    x = bash_app(stdout="{0}.out".format(fname))
    print("Waiting ....")
    print(x.result())


if __name__ == "__main__":

    test_python()
    test_bash()



"""Testing bash apps
"""
import parsl
from parsl import *

import argparse

# parsl.set_stream_logger()

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def increment(x):
    return x + 1


@App('python', dfk)
def slow_increment(x, dur):
    import time
    time.sleep(dur)
    return x + 1


def test_increment(depth=5):
    """Test simple pipeline A->B...->N
    """
    futs = {0: 0}
    for i in range(1, depth):
        futs[i] = increment(futs[i - 1])

    print([futs[i].result() for i in futs if not isinstance(futs[i], int)])


def test_increment_slow(depth=4):
    """Test simple pipeline A->B...->N with delay
    """
    futs = {0: 0}
    for i in range(1, depth):
        futs[i] = slow_increment(futs[i - 1], 0.5)

    print(futs[i])
    print([futs[i].result() for i in futs if not isinstance(futs[i], int)])


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-w", "--width", default="5",
                        help="width of the pipeline")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    # test_increment(depth=int(args.width))
    test_increment_slow(depth=int(args.width))



import os
from glob import glob
import logging

logger = logging.getLogger(__name__)


def make_rundir(config=None, path=None):
    """When a path has not been specified, make the run directory.

    Creates a rundir with the following hierarchy:
        ./runinfo <- Home of all run directories
          |----000
          |----001 <- Directories for each run
          | ....
          |----NNN

    Kwargs:
        - path (str): String path to a specific run dir
               Default : None.
    """
    try:
        if not path:
            path = None
        elif config["globals"].get('runDir', None):
            path = config["globals"]['runDir']

        if not path:
            path = "./runinfo"

        if not os.path.exists(path):
            os.makedirs(path)

        prev_rundirs = glob(os.path.join(path, "[0-9]*"))

        current_rundir = os.path.join(path, '000')

        if prev_rundirs:
            # Since we globbed on files named as 0-9
            x = sorted([int(os.path.basename(x)) for x in prev_rundirs])[-1]
            current_rundir = os.path.join(path, '{0:03}'.format(x + 1))

        os.makedirs(current_rundir)
        logger.debug("Parsl run initializing in rundir:{0}".format(current_rundir))
        return os.path.abspath(current_rundir)

    except Exception as e:
        logger.error("Failed to create a run directory")
        logger.error("Error: {0}".format(e))
        exit(-1)



localThreads = {
    "sites": [
        {"site": "Local_Threads",
         "auth": {"channel": None},
         "execution": {
             "executor": "threads",
             "provider": None,
             "maxThreads": 4
         }
         }],
    "globals": {"lazyErrors": True}
}

localIPP = {
    "sites": [
        {"site": "Local_IPP",
         "auth": {
             "channel": None,
         },
         "execution": {
             "executor": "ipp",
             "provider": "local",  # LIKELY SHOULD BE BOUND TO SITE
             "block": {  # Definition of a block
                 "taskBlocks": 4,       # total tasks in a block
                 "initBlocks": 4,
             }
         }
         }]
}

"""Use the following config with caution.
"""

localIPPReuse = {
    "sites": [
        {"site": "Local_IPP",
         "auth": {
             "channel": None,
         },
         "execution": {
             "executor": "ipp",
             "provider": "local",  # LIKELY SHOULD BE BOUND TO SITE
             "script_dir": ".scripts",
             "block": {  # Definition of a block
                 "nodes": 1,            # of nodes in that block
                 "taskBlocks": 1,       # total tasks in a block
                 "initBlocks": 1,
                 "maxBlocks": 1,
             }
         }
         }
    ],
    "globals": {
        "lazyErrors": True
    },
    "controller": {"reuse": True}
}



class ProviderFactoryError(Exception):
    """Base class for all exceptions.

    Only to be invoked when only a more specific error is not available.
    """

    def __repr__(self):
        return "Pool:{0}, Reason:{1}".format(self.pool, self.reason)

    def __str__(self):
        return self.__repr__()


class BadConfig(ProviderFactoryError):
    """The user provided a bad config.

    Contains:
        - Reason
        - Pool name
    """

    def __init__(self, pool, reason):
        super().__init__()
        self.pool = pool
        self.reason = reason



import parsl
from parsl import *
import time
import argparse

# parsl.set_stream_logger()
config = {
    "sites": [
        {"site": "Local_Threads",
         "auth": {"channel": None},
         "execution": {
             "executor": "threads",
             "provider": None,
             "maxThreads": 4,
         }
         }],
    "globals": {"lazyErrors": True,
    }
}

dfk = DataFlowKernel(config=config)


@App('python', dfk, cache=True)
def random_uuid(x, cache=True):
    import uuid
    return str(uuid.uuid4())


def test_python_memoization(n=4):
    """Testing python memoization disable
    """
    x = random_uuid(0)
    print(x.result())

    for i in range(0, n):
        foo = random_uuid(0)
        print(foo.result())
        assert foo.result() == x.result(), "Memoized results were not used"


@App('bash', dfk, cache=True)
def slow_echo_to_file(msg, outputs=[], stderr='std.err', stdout='std.out'):
    return 'sleep 1; echo {0} > {outputs[0]}'


def test_bash_memoization(n=4):
    """Testing bash memoization
    """

    print("Launching : ", n)
    x = slow_echo_to_file("hello world", outputs=['h1.out'])
    x.result()

    start = time.time()
    d = {}
    for i in range(0, n):
        d[i] = slow_echo_to_file("hello world", outputs=['h1.out'])

    print("Waiting for results from round1")
    [d[i].result() for i in d]
    end = time.time()
    delta = end - start
    assert delta < 0.1, "Memoized results were not used"


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    x = test_python_memoization(n=4)
    x = test_bash_memoization(n=4)



import parsl
import time
from parsl.execution_provider.local.local import Local
parsl.set_stream_logger()


def test_config_A():

    local = Local(config={'site': 'Local'})

    x = local.submit("ipengine ", blocksize=1)
    print("Started : ", x)
    local.status([x])
    time.sleep(3)
    y = local.submit("ipengine ", blocksize=1)
    print("Started : ", y)
    time.sleep(3)

    print(local.cancel([x, y]))
    print(local.status([x, y]))


if __name__ == "__main__":

    test_config_A()
    # test_config_B
    # test_config_C



from .canning import (
    Reference, can_map, uncan_map, can, uncan,
    use_dill, use_cloudpickle, use_pickle,
)
from .serialize import (
    serialize_object, deserialize_object,
    pack_apply_message, unpack_apply_message,
)

__all__ = (
    'Reference',
    'can_map',
    'uncan_map',
    'can',
    'uncan',
    'use_dill',
    'use_cloudpickle',
    'use_pickle',
    'serialize_object',
    'deserialize_object',
    'pack_apply_message',
    'unpack_apply_message',
)



"""Definitions for the @App decorator and the APP classes.

The APP class encapsulates a generic leaf task that can be executed asynchronously.

"""
import logging
from inspect import signature, Parameter

# Logging moved here in the PEP8 conformance fixes.
logger = logging.getLogger(__name__)


class AppBase (object):
    """This is the base class that defines the two external facing functions that an App must define.

    The  __init__ () which is called when the interpretor sees the definition of the decorated
    function, and the __call__ () which is invoked when a decorated function is called by the user.

    """

    def __init__(self, func, executor, walltime=60, sites='all', cache=False, exec_type="bash"):
        """Constructor for the APP object.

        Args:
             - func (function): Takes the function to be made into an App
             - executor (executor): Executor for the execution resource

        Kwargs:
             - walltime (int) : Walltime in seconds for the app execution
             - sites (str|list) : List of site names that this app could execute over. default is 'all'
             - exec_type (string) : App type (bash|python)
             - cache (Bool) : Enable caching of this app ?

        Returns:
             - APP object.

        """
        self.__name__ = func.__name__
        self.func = func
        self.executor = executor
        self.exec_type = exec_type
        self.status = 'created'
        self.sites = sites
        self.cache = cache

        sig = signature(func)
        self.kwargs = {}
        for s in sig.parameters:
            if sig.parameters[s].default != Parameter.empty:
                self.kwargs[s] = sig.parameters[s].default

        self.stdout = sig.parameters['stdout'].default if 'stdout' in sig.parameters else None
        self.stderr = sig.parameters['stderr'].default if 'stderr' in sig.parameters else None
        self.inputs = sig.parameters['inputs'].default if 'inputs' in sig.parameters else []
        self.outputs = sig.parameters['outputs'].default if 'outputs' in sig.parameters else []

    def __call__(self, *args, **kwargs):
        """The __call__ function must be implemented in the subclasses."""
        raise NotImplementedError


def app_wrapper(func):

    def wrapper(*args, **kwargs):
        logger.debug("App wrapper begins")
        x = func(*args, **kwargs)
        logger.debug("App wrapper ends")
        return x

    return wrapper


def App(apptype, executor, walltime=60, cache=False, sites='all'):
    """The App decorator function.

    Args:
        - apptype (string) : Apptype can be bash|python
        - executor (Executor) : Executor object wrapping threads/process pools etc.

    Kwargs:
        - walltime (int) : Walltime for app in seconds,
             default=60
        - sites (str|List) : List of site names on which the app could execute
             default='all'
        - cache (Bool) : Enable caching of the app call
             default=False

    Returns:
         An AppFactory object, which when called runs the apps through the executor.
    """
    from parsl import APP_FACTORY_FACTORY

    def Exec(f):
        return APP_FACTORY_FACTORY.make(apptype, executor, f,
                                        sites=sites,
                                        cache=cache,
                                        walltime=walltime)

    return Exec



"""Testing bash apps
"""
import parsl
from parsl import *

import argparse

# parsl.set_stream_logger()

workers = ThreadPoolExecutor(max_workers=100)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def random():
    import random
    return random.randint(1, 10)


@App('python', dfk)
def slow_increment(x, dur):
    import time
    time.sleep(dur)
    return x + 1


@App('python', dfk)
def join(inputs=[]):
    return sum(inputs)


@App('python', dfk)
def test_xor_split():
    """Test XOR split. Do A if x else B
    """
    x = random()

    if x.result() > 5:
        print("Result > 5")
        return 0
    else:
        print("Result < 5")
        return 1


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-w", "--width", default="10",
                        help="width of the pipeline")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    futures = {}

    for i in range(int(args.width)):
        futures[i] = test_xor_split()

    print([futures[key].result() for key in futures])



# A point in the workflow process where multiple paths converge into one single thread with
# synchronization

from parsl import *
import argparse

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def add(x, y):
    return x + y


@App('python', dfk)
def subtract(x, y):
    return y - x


@App('python', dfk)
def square(x, y):
    return (x + y)**2


@App('python', dfk)
def double(x, y):
    return 2 * (x + y)


@App('python', dfk)
def sum_elements(n=[]):
    s = 0
    for i in range(len(n)):
        s += n[i]
    return s


def test_or_split(x=4, y=5):
    num = []
    if x < 5:
        num.append(add(x, y).result())
    if y > 7:
        num.append(subtract(x, y).result())
    if x >= 5:
        num.append(square(x, y).result())
    if y <= 7:
        num.append(double(x, y).result())
    print(sum_elements(num).result())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", "--num1", default="4",
                        action="store", dest="x", type=int)
    parser.add_argument("-y", "--num2", default="5",
                        action="store", dest="y", type=int)
    args = parser.parse_args()
    test_or_split(args.x, args.y)



import zmq

context = zmq.Context()


class ZmQueue():

    def __init__(self, incoming, outgoing):

        self.incoming = context.socket(zmq.PULL)
        self.incoming.connect(incoming)

        self.outgoing = context.socket(zmq.PUSH)
        self.outgoing.connect(outgoing)

    def get(self):
        print("Waiting on ", self.incoming)
        x = self.incoming.recv()
        print("Received message ", x)
        return x

    def put(self, msg):
        print("Posting on ", self.outgoing)
        print("Sending message", msg)
        return self.outgoing.send(msg)



import logging
import json
import globus_sdk
# from globus_sdk.exc import TransferAPIError


logger = logging.getLogger(__name__)

'''
'Parsl Application' OAuth2 client registered with Globus Auth
by lukasz@globusid.org
'''
CLIENT_ID = '8b8060fd-610e-4a74-885e-1051c71ad473'
REDIRECT_URI = 'https://auth.globus.org/v2/web/auth-code'
SCOPES = ('openid '
          'urn:globus:auth:scope:transfer.api.globus.org:all')

TOKEN_FILE = 'runinfo/.globus.json'


get_input = getattr(__builtins__, 'raw_input', input)


def _load_tokens_from_file(filepath):
    with open(filepath, 'r') as f:
        tokens = json.load(f)
    return tokens


def _save_tokens_to_file(filepath, tokens):
    with open(filepath, 'w') as f:
        json.dump(tokens, f)


def _update_tokens_file_on_refresh(token_response):
    _save_tokens_to_file(TOKEN_FILE, token_response.by_resource_server)


def _do_native_app_authentication(client_id, redirect_uri,
                                  requested_scopes=None):

    client = globus_sdk.NativeAppAuthClient(client_id=client_id)
    client.oauth2_start_flow(
        requested_scopes=requested_scopes,
        redirect_uri=redirect_uri,
        refresh_tokens=True)

    url = client.oauth2_get_authorize_url()
    print('Native App Authorization URL: \n{}'.format(url))
    auth_code = get_input('Enter the auth code: ').strip()
    token_response = client.oauth2_exchange_code_for_tokens(auth_code)
    return token_response.by_resource_server


def _get_native_app_authorizer(client_id):
    tokens = None
    try:
        tokens = _load_tokens_from_file(TOKEN_FILE)
    except Exception:
        pass

    if not tokens:
        tokens = _do_native_app_authentication(
            client_id=client_id,
            redirect_uri=REDIRECT_URI,
            requested_scopes=SCOPES)
        try:
            _save_tokens_to_file(TOKEN_FILE, tokens)
        except Exception:
            pass

    transfer_tokens = tokens['transfer.api.globus.org']

    auth_client = globus_sdk.NativeAppAuthClient(client_id=client_id)

    return globus_sdk.RefreshTokenAuthorizer(
        transfer_tokens['refresh_token'],
        auth_client,
        access_token=transfer_tokens['access_token'],
        expires_at=transfer_tokens['expires_at_seconds'],
        on_refresh=_update_tokens_file_on_refresh)


def get_globus():
    Globus.init()
    return Globus()


"""
All communication with the Globus Auth and Globus Transfer services is enclosed
in the Globus class. In particular, the Globus class is reponsible for:
 - managing an OAuth2 authorizer - getting access and refresh tokens,
   refreshing an access token, storing to and retrieving tokens from
   .globus.json file,
 - submitting file transfers,
 - monitoring transfers.
"""


class Globus(object):

    authorizer = None

    @classmethod
    def init(cls):
        if cls.authorizer:
            return
        cls.authorizer = _get_native_app_authorizer(CLIENT_ID)

    @classmethod
    def get_authorizer(cls):
        return cls.authorizer

    @classmethod
    def transfer_file(cls, src_ep, dst_ep, src_path, dst_path):
        tc = globus_sdk.TransferClient(authorizer=cls.authorizer)
        td = globus_sdk.TransferData(tc, src_ep, dst_ep)
        td.add_item(src_path, dst_path)
        task = tc.submit_transfer(td)
        tc.task_wait(task['task_id'])
        task = tc.get_task(task['task_id'])
        if task['status'] != 'SUCCEEDED':
            logger.error(task)
            raise Exception('Transfer {}, from {}{} to {}{} to failed due to error: {}'.format(
                task['task_id'], src_ep, src_path, dst_ep, dst_path,
                task['nice_status_short_description']))



"""Testing python apps
"""

import parsl
from parsl import *
import argparse

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers], appCache=False)


@App('python', dfk)
def random_uuid(x):
    import uuid
    return str(uuid.uuid4())


def test_python_memoization(n=4):
    """Testing python memoization disable via DFK call
    """
    x = random_uuid(0)

    for i in range(0, n):
        foo = random_uuid(0)
        assert foo.result() != x.result(
        ), "Memoized results were used when memoization was disabled"


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    x = test_python_memoization(n=4)



"""Exceptions raise by Apps."""


class ParslError(Exception):
    """Base class for all exceptions.

    Only to be invoked when a more specific error is not available.
    """


class NotFutureError(ParslError):
    """A non future item was passed to a function that expected a future.

    This is basically a type error.
    """


class InvalidAppTypeError(ParslError):
    """An invalid app type was requested from the @App decorator."""


class AppException(ParslError):
    """An error raised during execution of an app.

    What this exception contains depends entirely on context
    """

    def __repr__(self):
        return "{0} Reason:{1}".format(self.__class__, self.reason)

    def __str__(self):
        return self.__repr__()


class AppBadFormatting(ParslError):
    """An error raised during formatting of a bash function.

    What this exception contains depends entirely on context
    Contains:
    reason(string)
    exitcode(int)
    retries(int/None)
    """

    def __init__(self, reason, exitcode, retries=None):
        super().__init__()
        self.reason = reason
        self.exitcode = exitcode
        self.retries = retries


class AppFailure(AppException):
    """An error raised during execution of an app.

    What this exception contains depends entirely on context
    Contains:
    reason(string)
    exitcode(int)
    retries(int/None)
    """

    def __init__(self, reason, exitcode, retries=None):
        super().__init__()
        self.reason = reason
        self.exitcode = exitcode
        self.retries = retries


class AppTimeout(AppException):
    """An error raised during execution of an app when it exceeds its allotted walltime.

    Contains:
    reason(string)
    exitcode(int)
    retries(int/None)
    """

    def __init__(self, reason, exitcode, retries=None):
        super().__init__()
        self.reason = reason
        self.exitcode = -55
        self.retries = retries


class BashAppNoReturn(AppException):
    """Bash app returned no string.

    Contains:
    reason(string)
    exitcode(int)
    retries(int/None)
    """

    def __init__(self, reason, exitcode, retries=None):
        super().__init__()
        self.reason = reason
        self.exitcode = -21
        self.retries = retries


class MissingOutputs(ParslError):
    """Error raised at the end of app execution due to missing output files.

    Contains:
    reason(string)
    outputs(List of strings/files..)
    """

    def __init__(self, reason, outputs):
        super().__init__()
        self.reason = reason
        self.outputs = outputs

    def __repr__(self):
        return "Missing Outputs: {0}, Reason:{1}".format(self.outputs, self.reason)

    def __str__(self):
        return "Reason:{0} Missing:{1}".format(self.reason, self.outputs)


class BadStdStreamFile(ParslError):
    """Error raised due to bad filepaths specified for STDOUT/ STDERR.

    Contains:
       reason(string)
       outputs(List of strings/files..)
       exception object
    """

    def __init__(self, outputs, exception):
        super().__init__()
        self._outputs = outputs
        self._exception = exception

    def __repr__(self):
        return "FilePath:[{}] Exception:{}".format(self._outputs,
                                                   self._exception)

    def __str__(self):
        return self.__repr__()


class DependencyError(ParslError):
    """Error raised at the end of app execution due to missing output files.

    Contains:
    reason(string)
    outputs(List of strings/files..)
    """

    def __init__(self, dependent_exceptions, reason, outputs):
        super().__init__()
        self.dependent_exceptions = dependent_exceptions
        self.reason = reason
        self.outputs = outputs

    def __repr__(self):
        return "Missing Outputs: {0}, Reason:{1}".format(self.outputs, self.reason)

    def __str__(self):
        return "Reason:{0} Missing:{1}".format(self.reason, self.outputs)



# A point in a process where two or more branches reconverge without synchronization

from parsl import *
import random
import argparse

workers = ThreadPoolExecutor(max_workers=20)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    x = random.randint(1, 10)
    print(x)
    return x


@App('python', dfk)
def square(x):
    return x**2


@App('python', dfk)
def cubed(x):
    return x**3


@App('python', dfk)
def increment(x):
    return x + 1


@App('python', dfk)
def add_two(x):
    return x + 2


@App('python', dfk)
def sum_elements(x, y):
    return x + y


def test_multi_merge():
    r = rand().result()
    num1 = cubed(add_two(square(r).result()).result())
    num2 = cubed(add_two(increment(r).result()).result())
    print(sum_elements(num1.result(), num2.result()).result())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", "--num1", default="4",
                        action="store", dest="x", type=int)
    args = parser.parse_args()
    test_multi_merge()



"""Testing bash apps
"""
import parsl
from parsl import *

print("Parsl version: ", parsl.__version__)


# parsl.set_stream_logger()
workers = ThreadPoolExecutor(max_workers=8)
dfk = DataFlowKernel(executors=[workers])

# Generate a random number


@App('python', dfk)
def generate(limit):
    from random import randint
    """Generate a random integer and return it"""
    return randint(1, limit)

# write a message to a file


@App('bash', dfk)
def save(message, outputs=[]):
    return 'echo %s &> {outputs[0]}' % (message)


def test_procedural(N=10):
    """Procedural workflow example from docs on
    Composing a workflow
    """
    message = generate(N)

    saved = save(message, outputs=['output.txt'])

    with open(saved.outputs[0].result().filepath, 'r') as f:
        item = int(f.read().strip())
        assert item <= N, "Expected file to contain int <= N"
        assert item >= 1, "Expected file to contain int >= 1"


if __name__ == "__main__":

    test_procedural()



"""Testing python apps
"""
import parsl
from parsl import *
import argparse
import json


def test_immutable_config(n=2):
    """Regression test for immutable config #98
    """

    localThreads = {
        "sites": [
            {"site": "Local_Threads",
             "auth": {"channel": None},
             "execution": {
                 "executor": "threads",
                 "provider": None,
                 "maxThreads": 4
             }
             }],
        "globals": {"lazyErrors": True}
    }

    original = json.dumps(localThreads, sort_keys=True)
    dfk = DataFlowKernel(config=localThreads)
    after = json.dumps(localThreads, sort_keys=True)

    dfk.cleanup()
    assert original == after, "Config modified"


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    x = test_immutable_config()



"""Testing early attach behavior with LoadBalanced view

Test setup:

Start the ipcontroller and 1 ipengine, and run this script.
The time to finish the 10 apps should be ~10s.

In the second run, start the parsl script, and as soon as the run starts,
start additional ipengines. The time to finish the 10 apps should still be ~10s.

This shows that the LoadBalanced View simply routes tasks to the available engines
at the time the apps were submitted to it. It is not capable of rebalancing the apps
among the engine once it has been sent to the the engine's queue.


"""
import parsl
from parsl import *

print("Parsl version: ", parsl.__version__)

import time

# parsl.set_stream_logger()

workers = IPyParallelExecutor()
dfk = DataFlowKernel(workers)


@App('python', dfk)
def sleep_double(x):
    import time
    time.sleep(1)
    return x * 2


if __name__ == "__main__":

    print("Starting launch")

    jobs = {}
    for i in range(0, 20):
        jobs[i] = sleep_double(i)

    start = time.time()
    for i in range(0, 10):
        print(jobs[i].result())
    print("Time to finish : ", time.time() - start)



"""Config which uses threads.

Threads are useful, for example, for local lightweight apps
and IPP workers for heavy weight applications.
The @app decorator in the example uses a new parameter sites=[<list of sites>]
to specify the site to which apps should be directed.

"""
threads_ipp = {
    "sites": [
        {"site": "Local_threads",
         "auth": {
             "channel": None,
         },
         "execution": {
             "executor": "threads",
             "provider": None,
             "maxThreads": 4
         }
         },
        {"site": "Local_IPP",
         "auth": {
             "channel": None,
         },
         "execution": {
             "executor": "ipp",
             "provider": "local",
             "block": {                 # Definition of a block
                 "nodes": 1,            # of nodes in that block
                 "taskBlocks": 1,       # total tasks in a block
                 "walltime": "00:05:00",
                 "initBlocks": 4,
             }
         }
         }],

    "globals": {
        "lazyErrors": True
    }
}


"""The following config uses two IPP sites designed for python apps which may
not show any performance improvements on local threads. This also allows you to
send work to two separate remote sites, or to two separate partitions.
"""

multi_ipp = {
    "sites": [
        {"site": "Local_IPP_1",
         "auth": {"channel": None},
         "execution": {
             "executor": "ipp",
             "provider": "local",
             "block": {
                 "nodes": 1,            # of nodes in that block
                 "taskBlocks": 1,       # total tasks in a block
                 "walltime": "00:15:00",
                 "initBlocks": 4,
             }
         }
         },
        {"site": "Local_IPP_2",
         "auth": {"channel": None},
         "execution": {
             "executor": "ipp",
             "provider": "local",
             "block": {
                 "nodes": 1,            # of nodes in that block
                 "taskBlocks": 1,       # total tasks in a block
                 "walltime": "00:15:00",
                 "initBlocks": 2,
             }
         }
         }],
    "globals": {
        "lazyErrors": True
    }
}



# In this workflow pattern, multiple instances of an activity can
# be created. Each time a new instance is created, a new thread is created that
# is independent of all the other threads

from parsl import *
import random
import argparse

workers = ThreadPoolExecutor(max_workers=10)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    x = random.randint(1, 10)
    return x


@App('python', dfk)
def square(x):
    return x**2


def test_multi_instances(x=5):
    numbers = []
    for i in range(x):
        numbers.append(square(rand().result()))
    print(numbers[i])
    print([numbers[i].result() for i in range(len(numbers))])


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", "--num", default="5",
                        action="store", dest="x", type=int)
    args = parser.parse_args()
    test_multi_instances(args.x)



"""Centralize app object creation."""

import logging
from inspect import getsource
from hashlib import md5
from inspect import signature
from parsl.app.bash_app import BashApp
from parsl.app.python_app import PythonApp
from parsl.app.errors import InvalidAppTypeError

logger = logging.getLogger(__name__)


class AppFactory(object):
    """AppFactory streamlines creation of apps."""

    def __init__(self, app_class, executor, func, cache=False, sites='all', walltime=60):
        """Construct an AppFactory for a particular app_class.

        Args:
            - app_class(Class) : An app class
            - executor(Executor) : An executor object which will handle app execution
            - func(Function) : The function to execute

        Kwargs:
            - walltime(int) : Walltime in seconds, default=60
            - sites (str|list) : List of site names that this app could execute over. default is 'all'
            - cache (Bool) : Enable caching of app.

        Returns:
            An AppFactory Object
        """
        self.__name__ = func.__name__
        self.app_class = app_class
        self.executor = executor
        self.func = func
        self.status = 'created'
        self.walltime = walltime
        self.sites = sites
        self.sig = signature(func)
        self.cache = cache
        # Function source hashing is done here to avoid redoing this every time
        # the app is called.
        if cache is True:
            try:
                fn_source = getsource(func)
            except OSError:
                logger.debug("Unable to get source code for AppCaching. Recommend creating module")
                fn_source = func.__name__

            self.func_hash = md5(fn_source.encode('utf-8')).hexdigest()
        else:
            self.func_hash = func.__name__

    def __call__(self, *args, **kwargs):
        """Create a new object of app_class with the args, execute the app_object and return the futures.

        Args:
             Arbitrary args to the decorated function

        Kwargs:
             Arbitrary kwargs to the decorated function

        Returns:
            (App_Future, [Data_Futures...])

        The call is mostly pass through
        """
        # Create and call the new App object
        app_obj = self.app_class(self.func,
                                 self.executor,
                                 sites=self.sites,
                                 walltime=self.walltime,
                                 cache=self.cache,
                                 fn_hash=self.func_hash)
        return app_obj(*args, **kwargs)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return '<class %s"%s for %s>' % (self.app_class.__name__,
                                         self.__class__.__name__,
                                         self.__name__)


class AppFactoryFactory(object):
    """An instance AppFactoryFactory will be factory that creates object of a particular kind.

    AppFactoryFactory has the various apps registered with it, and it will return an AppFactory
    that constructs objects of a specific kind.


    """

    def __init__(self, name):
        """Constructor.

        Args:
             name(string) : Name for the appfactory

        Returns:
             object(AppFactoryFactory)
        """
        self.name = name
        self.apps = {'bash': BashApp,
                     'python': PythonApp}

    def make(self, kind, executor, func, **kwargs):
        """Creates a new App of the kind specified.

        Args:
            kind(string) : For now only(bash|python)
            executor(Executor) : An executor object which will handle app execution
            func(Function) : The function to execute

        Kwargs:
            Walltime(int) : Walltime in seconds
            Arbritrary kwargs passed onto the AppFactory

        Raises:
            InvalidAppTypeError

        Returns:
            An AppFactory object bound to the specific app_class kind

        """
        if kind in self.apps:
            return AppFactory(self.apps[kind],
                              executor,
                              func,
                              **kwargs)

        else:
            logger.error("AppFactory:%s Invalid app kind requested : %s ",
                         self.name, kind)
            raise InvalidAppTypeError(
                "AppFactory:%s Invalid app kind requested : %s ",
                self.name, kind)



from parsl import *
import parsl
import libsubmit

print(parsl.__version__)
print(libsubmit.__version__)

# parsl.set_stream_logger()

from .local import localIPP
dfk = DataFlowKernel(config=localIPP)


@App("bash", dfk)
def bash_app(stdout=None, stderr=None):
    return 'echo "Hello from $(uname -a)" ; sleep 2'


def test_bash():
    """Testing basic scaling|Bash 0 -> 1 block """

    import os
    fname = os.path.basename(__file__)

    x = bash_app(stdout="{0}.out".format(fname))
    print("Waiting ....")
    print(x.result())



# In this workflow, there is a point in the process where multiple subprocesses/activities
# converge into one single thread of control

from parsl import *
import argparse

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def square(x):
    return x**2


@App('python', dfk)
def increment(x):
    return x + 1


@App('python', dfk)
def join(x, y):
    return x + y


def test_join(x=5):
    i = []
    for j in range(x):
        a = increment(j).result()
        b = square(j).result()
        i.append(join(a, b).result())
    total = sum(i)
    print(total)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--num", default="5",
                        action="store", dest="a", type=int)
    args = parser.parse_args()
    test_join(args.a)



"""Testing bash apps
"""
import parsl
from parsl import *
import time
import argparse

# parsl.set_stream_logger()
workers = IPyParallelExecutor()
print("Using ipyparallel workers")
# workers = ThreadPoolExecutor(max_workers=4)
print("Using threads pool 4")
dfk = DataFlowKernel(workers)


@App('python', dfk)
def double(x):
    return x * 2


def plain_double(x):
    return x * 2


def test_plain(n=10):
    start = time.time()
    x = []
    for i in range(0, n):
        x.extend([plain_double(i)])

    print(sum(x))

    ttc = time.time() - start
    print("Total time : ", ttc)

    return ttc


def test_parallel(n=10):
    start = time.time()
    x = []
    for i in range(0, n):
        x.extend([double(i)])

    print(sum([i.result() for i in x]))

    ttc = time.time() - start
    print("Total time : ", ttc)

    return ttc


def test_parallel2(n=10):
    start = time.time()
    x = []
    for i in range(0, n):
        x.extend([double(i)])

    ttc = time.time() - start
    print("Total time : ", ttc)

    return ttc


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    x = test_plain(int(args.count))
    # x = test_parallel(int(args.count))
    x = test_parallel2(int(args.count))

    # x = test_stdout()
    # raise_error(0)



"""Home of the configuration defaults."""

import copy
import collections
import pprint
import logging

logger = logging.getLogger(__name__)


def pp_config(config):
    """Pretty print the config.

    This should be part of the default logging to the debug logs.

    Args:
        - config (dict) : The config data structure
    """
    logger.debug(pprint.pformat(config, indent=2))
    return


def recursive_update(template, userdata):
    """Recursively update the template with userdata.

    If we don't do this the value updates for nested collections
    would get simply overwritten rathen than updated.

    Args:
        template (dict) : The templated dict
        userdata (dict) : User supplied config dict structure

    Returns:
        Updated template stucture.
    """
    for k, v in userdata.items():
        if isinstance(v, collections.Mapping):
            template[k] = recursive_update(template.get(k, {}), v)
        else:
            template[k] = v

    return template


def update_config(config, rundir):
    """Update the config datastructure with defaults.

    This is the one centralized location where the default live.

    Args:
         - config (dict) : The config dictionary

    Returns:
         - A standardized config dict if a config was passed, else None.
    """
    if not config:
        return None

    config_base = {"sites": [],
                   "globals": {
                       "lazyErrors": True,   # Bool
                       "usageTracking": True,  # Bool
                       "strategy": 'simple',  # ('simple',...)
                       "appCache": True,  # Bool
                       "checkpoint": None,  # (None, 'periodic', 'dfk_exit', 'task_exit')
                       "checkpointPeriod": None,  # (None, 'HH:MM:SS')
                       "checkpointFiles": None,   # List of checkpoint files
                   },
                   "controller": {
                       "mode": "auto"
                   }
    }

    _config = copy.deepcopy(config)
    sites = copy.deepcopy(_config["sites"])
    del _config["sites"]

    recursive_update(config_base, _config)

    config_sites = []
    for site in sites:
        site_base = {
            "auth": {},
            "execution": {
                "scriptDir": rundir,
                "block": {
                    "nodes": 1,
                    "taskBlocks": 1,
                    "minBlocks": 0,
                    "initBlocks": 0,
                    "maxBlocks": 10,
                    "parallelism": 0.75,
                    "walltime": "00:20:00",
                    "options": {
                    }
                }
            }
        }

        recursive_update(site_base, site)
        config_sites.extend([site_base])

    config_base["sites"] = config_sites

    pp_config(config_base)

    return config_base



# In this pattern, multiple instances of an activity can be enabled, with the
# number of instances of a given activity known during runtime before the
# instances of the activity are created

from parsl import *
import random
import argparse

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    x = random.randint(1, 10)
    return x


@App('python', dfk)
def square(x):
    return x**2


@App('python', dfk)
def increment_one(x):
    return x + 1


@App('python', dfk)
def sum_elements(x=[]):
    total = 0
    for i in range(len(x)):
        total += x[i].result()
    return total


def test_multi_instances(x=5):
    numbers = []
    [numbers.append(increment_one(square(rand().result()).result()))
     for i in range(x)]
    total = sum_elements(numbers)
    print(total.result())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", "--int", default="5",
                        action="store", dest="x", type=int)
    args = parser.parse_args()
    test_multi_instances(args.x)



# A given process is terminated once there are no longer any active activities
# in the workflow

from parsl import *
import random
import argparse
import time

workers = ThreadPoolExecutor(max_workers=10)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    x = random.randint(1, 10)
    return x


@App('python', dfk)
def square(x):
    return x**2


@App('python', dfk)
def increment(x):
    return x + 1


@App('python', dfk)
def increment_slow(x):
    time.sleep(5)
    return x + 1


@App('python', dfk)
def sum_elements(x, y):
    return x + y


def test_implicit_termination(x=5):
    numbers = []
    numbers.append(increment_slow(rand().result()))
    for i in range(x):
        y = rand().result()
        numbers.append(square(increment(y).result()).result())
    print(numbers[0].result())
    return


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", "--int", default="5",
                        action="store", dest="x", type=int)
    args = parser.parse_args()
    test_implicit_termination(args.x)



class ExecutorException(Exception):
    """Base class for all exceptions.

    Only to be invoked when only a more specific error is not available.
    """


class TaskExecException(ExecutorException):
    """Task execution raised an error in the remote process."""



import os

USERNAME = os.environ["COOLEY_USERNAME"]

singleNodeLocal = {
    "sites": [{
        "site": "ALCF_Cooley_Remote",
        "auth": {
            "channel": "local",
            "hostname": "cooleylogin1.alcf.anl.gov",
            "username": USERNAME,
            "scriptDir": "/home/{}/parsl_scripts/".format(USERNAME)
        },
        "execution": {
            "executor": "ipp",
            "provider": "cobalt",
            "block": {      # Definition of a block
                "nodes": 1,            # of nodes in that block
                "taskBlocks": 1,       # total tasks in a block
                "walltime": "00:05:00",
                "initBlocks": 1,
                "maxBlocks": 1,
                "options": {
                    "partition": "debug",
                    "account": 'CSC249ADCD01',
                    "overrides": "source /home/yadunand/setup_cooley_env.sh"
                }
            }
        }
    }
    ],
    "controller": {"publicIp": "10.230.100.209"},
    "globals": {"lazyErrors": True}

}


singleNode = {
    "sites": [{
        "site": "ALCF_Cooley_Remote",
        "auth": {
            "channel": "ssh-il",
            "hostname": "cooleylogin1.alcf.anl.gov",
            "username": USERNAME,
            "scriptDir": "/home/{}/parsl_scripts/".format(USERNAME)
        },
        "execution": {
            "executor": "ipp",
            "provider": "cobalt",
            "block": {      # Definition of a block
                "nodes": 1,            # of nodes in that block
                "taskBlocks": 1,       # total tasks in a block
                "walltime": "00:05:00",
                "initBlocks": 1,
                "maxBlocks": 1,
                "options": {
                    "partition": "debug",
                    "overrides": "source /home/yadunand/setup_cooley_env.sh"
                }
            }
        }
    }
    ],
    "globals": {
        "lazyErrors": True
    }
}

multiNode = {
    "sites": [{
        "site": "ALCF_Cooley_Remote",
        "auth": {
            "channel": "ssh-il",
            "hostname": "cooleylogin1.alcf.anl.gov",
            "username": USERNAME,
            "scriptDir": "/home/{}/parsl_scripts/".format(USERNAME)
        },
        "execution": {
            "executor": "ipp",
            "provider": "cobalt",  # LIKELY SHOULD BE BOUND TO SITE
            "block": {  # Definition of a block
                "nodes": 1,            # of nodes in that block
                "taskBlocks": 1,       # total tasks in a block
                "initBlocks": 2,
                "maxBlocks": 2,
                "options": {
                    "partition": "debug",
                    "overrides": "source /home/yadunand/setup_cooley_env.sh"
                }
            }
        }
    }
    ],
    "globals": {
        "lazyErrors": True
    }
}



import os

if str(os.environ.get('PARSL_TESTING', False)).lower() != 'true':
    raise RuntimeError("must first run 'export PARSL_TESTING=True'")


def setup_package():
    import subprocess
    import time
    proc = subprocess.Popen(["ipcluster", "start", "-n", "4"])
    time.sleep(2)
    print("Started ipcluster with pid:{0}".format(proc))
    return proc


def teardown_package():
    import subprocess
    import time
    proc = subprocess.Popen(["ipcluster", "stop"])
    print("Stopping ipcluster")
    time.sleep(2)
    return proc



"""Testing bash apps
"""
import parsl
from parsl import *

print("Parsl version: ", parsl.__version__)


# parsl.set_stream_logger()
workers = ThreadPoolExecutor(max_workers=8)

# workers = ProcessPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('bash', dfk)
def echo_to_file(inputs=[], outputs=[], stderr='std.err', stdout='std.out', walltime=0.5):
    return """echo "sleeping";
    sleep 1 """


def test_walltime():
    """Testing walltime exceeded exception """
    x = echo_to_file()

    try:
        r = x.result()
        print("Got result : ", r)
    except Exception as e:
        print(e.__class__)
        print("Caught exception ", e)


if __name__ == "__main__":
    test_walltime()



"""Testing bash apps
"""
import parsl
from parsl import *

import time
import argparse

workers = ThreadPoolExecutor(max_workers=100)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def increment(x):
    return x + 1


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--count", default="100",
                        help="width of the pipeline")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    start = time.time()
    x = {}
    for i in range(int(args.count)):
        x[i] = increment(i)
    end = time.time()
    print("Launched {0} tasks in {1} s".format(args.count, end - start))



"""Parsl is a Parallel Scripting Library, designed to enable efficient workflow execution.

Importing
---------

To get all the required functionality, we suggest importing the library as follows:

>>> import parsl
>>> from parsl import *

Logging
-------

Following the general logging philosophy of python libraries, by default
`Parsl <https://github.com/swift-lang/swift-e-lab/>`_ doesn't log anything.
However the following helper functions are provided for logging:

1. set_stream_logger
    This sets the logger to the StreamHandler. This is quite useful when working from
    a Jupyter notebook.

2. set_file_logger
    This sets the logging to a file. This is ideal for reporting issues to the dev team.

"""
import logging

from parsl.version import VERSION
from parsl.app.app import App
from parsl.executors.threads import ThreadPoolExecutor
from parsl.executors.ipp import IPyParallelExecutor
from parsl.data_provider.files import File

from parsl.dataflow.dflow import DataFlowKernel
from parsl.app.app_factory import AppFactoryFactory
APP_FACTORY_FACTORY = AppFactoryFactory('central')

__author__ = 'Yadu Nand Babuji'
__version__ = VERSION

__all__ = [
    'App', 'DataFlowKernel', 'File', 'ThreadPoolExecutor',
    'IPyParallelExecutor', 'set_stream_logger', 'set_file_logger'
]


def set_stream_logger(name='parsl', level=logging.DEBUG, format_string=None):
    """Add a stream log handler.

    Args:
         - name (string) : Set the logger name.
         - level (logging.LEVEL) : Set to logging.DEBUG by default.
         - format_string (sting) : Set to None by default.

    Returns:
         - None
    """
    if format_string is None:
        # format_string = "%(asctime)s %(name)s [%(levelname)s] Thread:%(thread)d %(message)s"
        format_string = "%(asctime)s %(name)s:%(lineno)d [%(levelname)s]  %(message)s"

    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(format_string, datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def set_file_logger(filename, name='parsl', level=logging.DEBUG, format_string=None):
    """Add a stream log handler.

    Args:
        - filename (string): Name of the file to write logs to
        - name (string): Logger name
        - level (logging.LEVEL): Set the logging level.
        - format_string (string): Set the format string

    Returns:
       -  None
    """
    if format_string is None:
        format_string = "%(asctime)s %(name)s:%(lineno)d [%(levelname)s]  %(message)s"

    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.FileHandler(filename)
    handler.setLevel(level)
    formatter = logging.Formatter(format_string, datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)


class NullHandler(logging.Handler):
    """Setup default logging to /dev/null since this is library."""

    def emit(self, record):
        pass


logging.getLogger('parsl').addHandler(NullHandler())



import logging
import sys
import concurrent.futures as cf
from parsl.executors.base import ParslExecutor

logger = logging.getLogger(__name__)


class ThreadPoolExecutor(ParslExecutor):
    """The thread pool executor."""

    def __init__(self, max_workers=2, thread_name_prefix='',
                 execution_provider=None, config=None, **kwargs):
        """Initialize the thread pool.

        Config options that are really used are :

        config.sites.site.execution.options = {"maxThreads" : <int>,
                                               "threadNamePrefix" : <string>}

        Kwargs:
           - max_workers (int) : Number of threads (Default=2) (keeping name workers/threads for backward compatibility)
           - thread_name_prefix (string) : Thread name prefix (Only supported in python v3.6+
           - execution_provider (ep object) : This is ignored here
           - config (dict): The config dict object for the site:


        """
        self._scaling_enabled = False
        if not config:
            config = {"execution": {}}
        if "maxThreads" not in config["execution"]:
            config["execution"]["maxThreads"] = max_workers
        if "threadNamePrefix" not in config["execution"]:
            config["execution"]["threadNamePrefix"] = thread_name_prefix

        self.config = config

        if sys.version_info > (3, 6):
            self.executor = cf.ThreadPoolExecutor(max_workers=config["execution"]["maxThreads"],
                                                  thread_name_prefix=config["execution"]["threadNamePrefix"])
        else:
            self.executor = cf.ThreadPoolExecutor(max_workers=config["execution"]["maxThreads"])

    @property
    def scaling_enabled(self):
        return self._scaling_enabled

    def submit(self, *args, **kwargs):
        """Submits work to the thread pool.

        This method is simply pass through and behaves like a submit call as described
        here `Python docs: <https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor>`_

        Returns:
              Future
        """
        return self.executor.submit(*args, **kwargs)

    def scale_out(self, workers=1):
        """Scales out the number of active workers by 1.

        This method is notImplemented for threads and will raise the error if called.

        Raises:
             NotImplemented exception
        """

        raise NotImplementedError

    def scale_in(self, workers=1):
        """Scale in the number of active workers by 1.

        This method is notImplemented for threads and will raise the error if called.

        Raises:
             NotImplemented exception
        """

        raise NotImplementedError

    def shutdown(self, block=False):
        """Shutdown the ThreadPool.

        This method is notImplemented for threads and will raise the error if called.
        The interface documentation for IPP is `here <http://ipyparallel.readthedocs.io/en/latest/api/ipyparallel.html#ipyparallel.Client.shutdown>`_

        Kwargs:
            - block (Bool): To block for confirmations or not

        """
        x = self.executor.shutdown(wait=block)
        logger.debug("Done with executor shutdown")
        return x



"""Exceptions raise by Executors."""


class ExecutorError(Exception):
    """Base class for all exceptions.

    Only to be invoked when only a more specific error is not available.
    """

    def __repr__(self):
        return "Site:{0}, Reason:{1}".format(self.site, self.reason)

    def __str__(self):
        return self.__repr__()


class ScalingFailed(ExecutorError):
    """Scaling failed due to error in Execution provider."""

    def __init__(self, sitename, reason):
        self.site = sitename
        self.reason = reason


class ControllerErr(ExecutorError):
    """Error raise by IPP controller."""

    def __init__(self, reason):
        self.reason = reason

    def __repr__(self):
        return "Controller init failed:Reason:{0}".format(self.reason)

    def __str__(self):
        return self.__repr__()



import parsl
from parsl import *
import os
import time
import argparse

# parsl.set_stream_logger()
config = {
    "sites": [
        {"site": "Local_Threads",
         "auth": {"channel": None},
         "execution": {
             "executor": "threads",
             "provider": None,
             "maxThreads": 2,
         }
         }],
    "globals": {"lazyErrors": True,
                }
}

dfk = DataFlowKernel(config=config)


@App('python', dfk, cache=True)
def slow_double(x, sleep_dur=1):
    import time
    time.sleep(sleep_dur)
    return x * 2


def test_initial_checkpoint_write(n=4):
    """1. Launch a few apps and write the checkpoint once a few have completed
    """

    d = {}
    time.time()
    print("Launching : ", n)
    for i in range(0, n):
        d[i] = slow_double(i)
    print("Done launching")

    for i in range(0, n):
        d[i].result()
    print("Done sleeping")
    cpt_dir = dfk.checkpoint()

    cptpath = cpt_dir + '/dfk.pkl'
    print("Path exists : ", os.path.exists(cptpath))
    assert os.path.exists(
        cptpath), "DFK checkpoint missing: {0}".format(cptpath)

    cptpath = cpt_dir + '/tasks.pkl'
    print("Path exists : ", os.path.exists(cptpath))
    assert os.path.exists(
        cptpath), "Tasks checkpoint missing: {0}".format(cptpath)

    return


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    x = test_initial_checkpoint_write(n=4)



# An activity is enabled only if a certain milestone has been reached
# and has not expired yet. For example, if there are three activities A, B, and C,
# A may only be enabled if B has been executed and C has not been executed

from parsl import *
import random
import argparse
import time

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    x = random.randint(1, 10)
    print(x)
    return x


@App('python', dfk)
def square(x):
    if x > 5:
        return x**2
    else:
        time.sleep(5)
        return x**2


@App('python', dfk)
def increment(x):
    time.sleep(1)
    return x + 1


@App('python', dfk)
def cubed(x):
    return x**3


def test_milestone():
    r = rand().result()
    i = increment(r)
    s = square(r)
    while s.done() is not True:
        if i.done() is True:
            print(cubed(r).result())
            return


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", "--input", default="4",
                        action="store", dest="x", type=int)
    args = parser.parse_args()
    test_milestone()



from parsl import *
import parsl
import libsubmit

print(parsl.__version__)
print(libsubmit.__version__)

from .local import localIPP
dfk = DataFlowKernel(config=localIPP)


@App("python", dfk)
def python_app():
    import platform
    return "Hello from {0}".format(platform.uname())


def test_python(N=2):
    """Testing basic scaling|Python 0 -> 1 block """

    results = {}
    for i in range(0, N):
        results[i] = python_app()

    print("Waiting ....")
    for i in range(0, N):
        print(results[0].result())


if __name__ == '__main__':

    parsl.set_stream_logger()
    test_python()



import os
import time
import logging
from ipyparallel import Client

from parsl.executors.base import ParslExecutor
from parsl.executors.errors import *

logger = logging.getLogger(__name__)


class IPyParallelExecutor(ParslExecutor):
    """The IPython Parallel executor.

    This executor allows us to take advantage of multiple processes running locally
    or remotely via  IPythonParallel's pilot execution system.

    .. note::
           Some deficiencies with this executor are:

               1. Ipengine's execute one task at a time. This means one engine per core
                  is necessary to exploit the full parallelism of a node.
               2. No notion of remaining walltime.
               3. Lack of throttling means tasks could be queued up on a worker.

    """

    def compose_launch_cmd(self, filepath, engine_dir):
        """Reads the json contents from filepath and uses that to compose the engine launch command.

        Args:
            filepath: Path to the engine file
            engine_dir : CWD for the engines .

        """
        self.engine_file = os.path.expanduser(filepath)

        engine_json = None
        try:
            with open(self.engine_file, 'r') as f:
                engine_json = f.read()

        except OSError as e:
            logger.error("Could not open engine_json : ", self.engine_file)
            raise e

        return """cd {0}
cat <<EOF > ipengine.json
{1}
EOF

mkdir -p '.ipengine_logs'
ipengine --file=ipengine.json &>> .ipengine_logs/$JOBNAME.log
""".format(engine_dir, engine_json)

    def __init__(self, execution_provider=None,
                 reuse_controller=True,
                 engine_json_file='~/.ipython/profile_default/security/ipcontroller-engine.json',
                 engine_dir='.',
                 controller=None,
                 config=None):
        """Initialize the IPyParallel pool. The initialization takes all relevant parameters via KWargs.

        .. note::

              If initBlocks > 0, and a scalable execution_provider is attached, then the provider
              will be initialized here.

        Args:
             - self

        KWargs:
             - execution_provider (ExecutionProvider object)
             - reuse_controller (Bool) : If True ipp executor will attempt to connect to an available
               controller. Default: True
             - engine_json_file (str): Path to json engine file that will be used to compose ipp launch
               commands at scaling events. Default : '~/.ipython/profile_default/security/ipcontroller-engine.json'
             - engine_dir (str) : Alternative to above, specify the engine_dir
             - config (dict). Default: '.'
        """
        self.controller = controller
        self.engine_file = engine_json_file
        self.client_file = None

        if self.controller:
            # Find the Client json
            self.client_file = self.controller.client_file
            self.engine_file = self.controller.engine_file

            if not os.path.exists(self.client_file):
                logger.debug("Waiting for {0}".format(self.client_file))

            sleep_dur = 20  # 20 seconds
            for i in range(0, int(sleep_dur / 0.2)):
                time.sleep(0.2)
                if os.path.exists(self.client_file):
                    break

            if not os.path.exists(self.client_file):
                raise Exception("Controller client file is missing at {0}".format(self.client_file))

        self.executor = Client(url_file=self.client_file)
        self.config = config
        self.sitename = config['site'] if config else 'Static_IPP'
        # NOTE: Copying the config here only partially fixes the issue. There needs to be
        # multiple controllers launched by the factory, and each must have different jsons.
        # There could be timing issues here,
        # local_engine_json = "{0}.{1}.engine.json".format(self.config["site"], int(time.time()))
        # copyfile(engine_json_file, local_engine_json)
        # if not os.path.exists(self.config["execution"]["script_dir"]):
        #    os.makedirs(self.config["execution"]["script_dir"])

        self.launch_cmd = self.compose_launch_cmd(self.engine_file, engine_dir)
        self.execution_provider = execution_provider
        self.engines = []

        if reuse_controller:
            # Reuse existing controller if one is available
            pass

        if execution_provider:
            self._scaling_enabled = True
            logger.debug("Starting IpyParallelExecutor with provider:%s", execution_provider)
            try:
                for i in range(self.config["execution"]["block"].get("initBlocks", 1)):
                    eng = self.execution_provider.submit(self.launch_cmd, 1)
                    logger.debug("Launched block : {0}:{1}".format(i, eng))
                    if not eng:
                        raise(ScalingFailed(self.execution_provider.sitename,
                                            "Ipp executor failed to scale via execution_provider"))
                    self.engines.extend([eng])

            except Exception as e:
                logger.error("Scaling out failed : %s", e)
                raise e

        else:
            self._scaling_enabled = False
            logger.debug("Starting IpyParallelExecutor with no provider")

        self.lb_view = self.executor.load_balanced_view()
        logger.debug("Starting executor")

    @property
    def scaling_enabled(self):
        return self._scaling_enabled

    def submit(self, *args, **kwargs):
        """Submits work to the thread pool.

        This method is simply pass through and behaves like a submit call as described
        here `Python docs: <https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor>`_

        Returns:
              Future
        """
        # logger.debug("Got args : %s,", args)
        # logger.debug("Got kwargs : %s,", kwargs)
        return self.lb_view.apply_async(*args, **kwargs)

    def scale_out(self, *args, **kwargs):
        """Scales out the number of active workers by 1.

        This method is notImplemented for threads and will raise the error if called.

        """
        if self.execution_provider:
            r = self.execution_provider.submit(self.launch_cmd, *args, **kwargs)
            self.engines.extend([r])
        else:
            logger.error("No execution provider available")
            r = None

        return r

    def scale_in(self, blocks, *args, **kwargs):
        """Scale in the number of active workers by 1.

        This method is notImplemented for threads and will raise the error if called.

        Raises:
             NotImplemented exception
        """
        status = dict(zip(self.engines, self.execution_provider.status(self.engines)))

        # This works for blocks=0
        to_kill = [engine for engine in status if status[engine] == "RUNNING"][:blocks]

        if self.execution_provider:
            r = self.execution_provider.cancel(to_kill, *args, **kwargs)
        else:
            logger.error("No execution provider available")
            r = None

        return r

    def status(self):
        """Returns the status of the executor via probing the execution providers."""
        if self.execution_provider:
            status = self.execution_provider.status(self.engines)

        else:
            status = []

        return status

    def shutdown(self, hub=True, targets='all', block=False):
        """Shutdown the executor, including all workers and controllers.

        The interface documentation for IPP is `here <http://ipyparallel.readthedocs.io/en/latest/api/ipyparallel.html#ipyparallel.Client.shutdown>`_

        Kwargs:
            - hub (Bool): Whether the hub should be shutdown, Default:True,
            - targets (list of ints| 'all'): List of engine id's to kill, Default:'all'
            - block (Bool): To block for confirmations or not

        Raises:
             NotImplemented exception
        """
        if self.controller:
            logger.debug("IPP:Shutdown sequence: Attempting controller kill")
            self.controller.close()

        # We do not actually do executor.shutdown because
        # this blocks even when requested to not block, killing the
        # controller is more effective although impolite.
        # x = self.executor.shutdown(targets=targets,
        #                           hub=hub,
        #                           block=block)

        logger.debug("Done with executor shutdown")
        return True

    def __repr__(self):
        return "<IPP Executor for site:{0}>".format(self.sitename)


if __name__ == "__main__":

    pool1_config = {"poolname": "pool1",
                    "queue": "foo"}



import os
import logging
import concurrent.futures as cf
from parsl.executors.base import ParslExecutor
from parsl.data_provider.globus import get_globus

logger = logging.getLogger(__name__)


class DataManager(ParslExecutor):
    """DataManager uses the familiar Executor interface, where staging tasks are submitted
    to it, and DataFutures are returned.
    """

    """
    In general a site where remote file is going to be stage in is unknown until
    Executer submits an app that depends on the file. However, in most practical
    cases, a site where an app ir executed and a file needs to be staged in is
    known. Such cases should be detected by DataManager to optimize file
    transfers. Possible cases are:
    1. Config defines one site only.
    2. Config defines several sites but all of the sites share the filesystem /
       use the same Globus endpoint.
    3. Config defines several sites with different Globus endpoints but a user
       specified explicitely that apps must be executed on a particular site.
    """

    default_data_manager = None

    @classmethod
    def get_data_manager(cls):
        return cls.default_data_manager

    def __init__(self, max_workers=10, config=None):
        """Initialize the DataManager.

        Kwargs:
           - max_workers (int) : Number of threads (Default=10)
           - config (dict): The config dict object for the site:

        """
        self._scaling_enabled = False
        self.executor = cf.ThreadPoolExecutor(max_workers=max_workers)

        self.config = config
        if not self.config:
            self.config = {"sites": []}
        self.files = []
        self.globus = None

        if not DataManager.get_data_manager():
            DataManager.default_data_manager = self

    def submit(self, *args, **kwargs):
        """Submit a staging request."""
        return self.executor.submit(*args, **kwargs)

    def scale_in(self, blocks, *args, **kwargs):
        pass

    def scale_out(self, *args, **kwargs):
        pass

    def shutdown(self, block=False):
        return self.executor.shutdown(wait=block)

    def scaling_enabled(self):
        return self._scaling_enabled

    def add_file(self, file):
        if file.scheme == 'globus':
            if not self.globus:
                self.globus = get_globus()
            # keep a list of all remote files for optimization purposes (TODO)
            self.files.append(file)
            self._set_local_path(file)

    def _set_local_path(self, file):
        site, globus_ep = self._get_globus_site()
        file.local_path = os.path.join(
                globus_ep['local_directory'],
                file.filename)

    def _get_globus_site(self, site_name=None):
        for s in self.config['sites']:
            if site_name is None or s['site'] == site_name:
                if 'data' in s:
                    if 'globus' in s['data']:
                        return (s['site'], s['data']['globus'])
        raise Exception('No site with a Globus endpoint defined')

    def stage_in(self, file, site_name=None):
        ''' The stage_in call transports the file from the site of origin
        to the site. The function returns DataFuture.

        Args:
            - self
            - file (File) - file to stage in
            - site_name (str) - a name of a site the file is going to be staged in to.
                                If the site argument is not specified for a file
                                with 'globus' scheme, the file will be staged in to
                                the first site with the "globus" key in a config.
        '''

        if file.scheme == 'file':
            site_name = None
        elif file.scheme == 'globus':
            site_name, globus_ep = self._get_globus_site(site_name)

        df = file.get_data_future(site_name)
        if df:
            return df

        if file.scheme == 'file':
            f = self.submit(self._file_transfer_in, file)
        elif file.scheme == 'globus':
            f = self.submit(self._globus_transfer_in, file, globus_ep)

        from parsl.app.futures import DataFuture

        df = DataFuture(f, file)
        file.set_data_future(df, site_name)
        return df

    def stage_out(self, file, site_name=None):
        ''' The stage_out call transports the file from local filesystem
        to the remote Globus endpoint. The function return DataFuture.

        Args:
            - self
            - file (File) - file to stage out
            - site_name (str) - a name of a site the file is going to be staged out from.
                                If the site argument is not specified for a file
                                with 'globus' scheme, the file will be staged in to
                                the first site with the "globus" key in a config.
        '''

        if file.scheme == 'file':
            site_name = None
            f = self.submit(self._file_transfer_out)
            return f
        if file.scheme == 'globus':
            site_name, globus_ep = self._get_globus_site(site_name)
            f = self.submit(self._globus_transfer_out, file, globus_ep)
            return f

    def _file_transfer_in(self, file):
        pass

    def _globus_transfer_in(self, file, globus_ep):
        file.local_path = os.path.join(
                globus_ep['local_directory'], file.filename)
        dst_path = os.path.join(
                globus_ep['endpoint_path'], file.filename)
        self.globus.transfer_file(
                file.netloc, globus_ep['endpoint_name'],
                file.path, dst_path)

    def _file_transfer_out(self, file):
        pass

    def _globus_transfer_out(self, file, globus_ep):
        src_path = os.path.join(
                globus_ep['endpoint_path'], file.filename)
        self.globus.transfer_file(
                globus_ep['endpoint_name'], file.netloc,
                src_path, file.path)



# A point in the workflow process where one of several branches is chosen baced on a decision or
# workflow control data

from parsl import *
import argparse
import random

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    x = random.randint(1, 10)
    return x


@App('python', dfk)
def eval_number(x):
    if x > 5:
        print("Larger than 5")
        return return_one().result()
    else:
        print("Less than or equal to 5")
        return return_zero().result()


@App('python', dfk)
def return_one():
    return 1


@App('python', dfk)
def return_zero():
    return 0


def test_XOR_split(x=4):
    num = []
    for i in range(x):
        num.append(eval_number(rand().result()))
    for i in range(len(num)):
        print(num[i].result())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--a", "--number", default="4",
                        action="store", dest="a", type=int)
    args = parser.parse_args()
    test_XOR_split(args.a)



import parsl

parsl.set_stream_logger()

from parsl.execution_provider.provider_factory import ExecProviderFactory


def test_factory_1():

    config = {
        "sites": [
            {"site": "RCC_Midway_Remote",
             "auth": {
                 "channel": "ssh",
                 "hostname": "swift.rcc.uchicago.edu",
                 "username": "yadunand"
             },
             "execution": {
                 "executor": "ipp",
                 "provider": "slurm",  # LIKELY SHOULD BE BOUND TO SITE
                 "scriptDir": ".scripts",
                 "block": {  # Definition of a block
                     "nodes": 1,            # of nodes in that block
                     "taskBlocks": 1,        # total tasks in a block
                     "walltime": "00:05:00",
                     "Options": {
                         "partition": "debug",
                         "account": "pi-wilde",
                         "overrides": "#SBATCH--constraint=haswell"
                     }
                 }
             }
             }
        ],
        "globals": {
            "lazyErrors": True
        }
    }

    epf = ExecProviderFactory()
    epf.make(config)


def test_factory_2():

    config = {"site": "ipp_local",
              "execution":
              {"executor": "ipp",
               "provider": "local",
               "channel": "None",
               "options":
               {"initParallelism": 2,
                "maxParallelism": 2,
                "minParallelism": 0,
                   "walltime": "00:25:00",
                }
               }}

    epf = ExecProviderFactory()
    epf.make(config)


def test_factory_3():

    config = {
        "sites": [
            {"site": "Local",
             "auth": {
                 "channel": None
             },
             "execution": {
                 "executor": "threads",
                 "provider": None,  # LIKELY SHOULD BE BOUND TO SITE
                 "maxThreads": 4
             }
             }
        ],
        "globals": {
            "lazyErrors": True
        }
    }

    epf = ExecProviderFactory()
    epf.make(config)


if __name__ == '__main__':

    # test_factory_1()
    # test_factory_2()
    test_factory_3()



import os
import subprocess
import time
import random
import logging
import signal

from parsl.executors.errors import *

logger = logging.getLogger(__name__)


class Controller(object):
    """Start and maintain a ipyparallel controller."""

    def __init__(self, publicIp=None, port=None, portRange="", reuse=False,
                 log=True, ipythonDir="~/.ipython", mode="auto", profile=None):
        """Initialize ipython controllers to the user specified configs.

        The specifig config sections that will be used by this are in the dict
        config["controller"]

        KWargs:
              - publicIp (string): internal_ip, specify if this would be difficult to autofind.
              - interfaces (string): interfaces for zero_mq to listen on
                     Default: "*"
              - port (int): port
                     Default: A random port between 50000, 60000
              - portRange (string): <string <port_min>,<port_max>,
                     Default: 50000 - 60000
              - reuse (bool): <Reuse an existing controller>
              - ipythonDir (string) : IPython directory for IPP to store config files.
                     This will be overriden by the auto controller start.
                     Default: "~/.ipython"
              - profile (str)  : Name of the profile/site from the sites in the config.
                     Default : None
              - mode (str) : If "auto" the default automatic behavior is maintained.
                             If "manual" the controller is assumed to be created by the user.
        """
        logger.debug("Starting ipcontroller, baseDir:%s" % ipythonDir)

        self.mode = mode
        self.range_min = 50000
        self.range_max = 60000
        self.reuse = reuse
        self.port = ''
        self.ipythonDir = None
        self.profile = 'default'
        ipp_basedir = ''
        reuse_string = ''
        profile_string = ''

        if mode == "manual":
            return
        if self.reuse:
            reuse_string = '--reuse'

        if ipythonDir:
            self.ipythonDir = os.path.abspath(os.path.expanduser(ipythonDir))
            ipp_basedir = '--ipython-dir={0}'.format(self.ipythonDir)

        if profile:
            # Profile string is set only if non-"default" profile is specified
            self.profile = profile
            profile_string = '--profile={0}'.format(profile)

        # If portRange is specified pick a random port from that range
        try:
            if portRange:
                self.range_min, self.range_max = map(int, portRange.split(','))
                self.port = '--port={0}'.format(random.randint(self.range_min, self.range_max))
        except Exception as e:
            msg = "Bad portRange. IPPController needs portA,portB. Got {0}. Error {1}".format(portRange,
                                                                                              e)
            logger.error(msg)
            raise ControllerErr(msg)

        # If port is specified use it, this will override the portRange
        if port:
            self.port = '--port={0}'.format(port)

        # We have a publicIp default always = None
        if publicIp:
            self.publicIp = '--location={0}'.format(publicIp)
        else:
            self.publicIp = ''

        if log:
            stdout = open(os.path.join(self.ipythonDir, "{0}.controller.out".format(self.profile)), 'w')
            stderr = open(os.path.join(self.ipythonDir, "{0}.controller.err".format(self.profile)), 'w')
        else:
            stdout = open(os.devnull, 'w')
            stderr = open(os.devnull, 'w')

        interfaces = '--ip=*'

        try:
            opts = ['ipcontroller', ipp_basedir, interfaces, profile_string,
                    reuse_string, self.port, self.publicIp]
            logger.debug("Start opts: %s" % opts)
            self.proc = subprocess.Popen(opts, stdout=stdout, stderr=stderr, preexec_fn=os.setsid)

        except FileNotFoundError as e:
            msg = "Could not find ipcontroller. Please make sure that ipyparallel is installed and available in your env"
            logger.error(msg)
            raise ControllerErr(msg)

        except Exception as e:
            msg = "IPPController failed to start: {0}".format(e)
            logger.error(msg)
            raise ControllerErr(msg)

        return

    @property
    def engine_file(self):
        """Specify path to the ipcontroller-engine.json file.

        This file is stored in in the ipython_dir/profile folders.

        Returns :
              - str, File path to engine file
        """
        return os.path.join(self.ipythonDir,
                            'profile_{0}'.format(self.profile),
                            'security/ipcontroller-engine.json')

    @property
    def client_file(self):
        """Specify path to the ipcontroller-client.json file.

        This file is stored in in the ipython_dir/profile folders.

        Returns :
              - str, File path to client file
        """
        return os.path.join(self.ipythonDir,
                            'profile_{0}'.format(self.profile),
                            'security/ipcontroller-client.json')

    def close(self):
        """Terminate the controller process and it's child processes.

        Args:
              - None
        """
        if self.reuse:
            logger.debug("Ipcontroller not shutting down: reuse enabled")
            return

        if self.mode == "manual":
            logger.debug("Ipcontroller not shutting down: Manual mode")
            return

        try:
            pgid = os.getpgid(self.proc.pid)
            os.killpg(pgid, signal.SIGTERM)
            time.sleep(0.2)
            os.killpg(pgid, signal.SIGKILL)
            try:
                self.proc.wait(timeout=1)
                x = self.proc.returncode
                logger.debug("Controller exited with {0}".format(x))
            except subprocess.TimeoutExpired:
                logger.warn("Ipcontroller process:{0} cleanup failed. May require manual cleanup".format(self.proc.pid))

        except Exception as e:
            logger.warn("Failed to kill the ipcontroller process[{0}]: {1}".format(self.proc.pid,
                                                                                   e))



from parsl import *
import parsl
import libsubmit

print(parsl.__version__)
print(libsubmit.__version__)

localIPP = {
    "sites": [
        {"site": "Local_IPP",
         "auth": {
             "channel": None,
         },
         "execution": {
             "executor": "ipp",
             "provider": "local",  # LIKELY SHOULD BE BOUND TO SITE
             "block": {  # Definition of a block
                 "taskBlocks": 4,       # total tasks in a block
                 "initBlocks": 0,
                 "minBlocks": 0,
                 "maxBlocks": 10,
                 "parallelism": 0,
             }
         }
         }]
}

dfk = DataFlowKernel(config=localIPP)


@App("python", dfk)
def python_app():
    import platform
    return "Hello from {0}".format(platform.uname())


def test_python(N=2):
    """No blocks provisioned if parallelism==0

    If I set initBlocks=0 and parallelism=0 I don't think any blocks will be provisioned.
    I tested and the script makes no progress. Perhaps we should catch this case and present an error to users.
    """

    results = {}
    for i in range(0, N):
        results[i] = python_app()

    print("Waiting ....")
    for i in range(0, N):
        print(results[0].result())


if __name__ == '__main__':

    parsl.set_stream_logger()
    test_python()



# A set of activities is executed in an arbitrary order. Each activity
# in the set is executed, and no two activities are active at the same
# time

from parsl import *
import random
import argparse
import time

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    x = random.randint(1, 10)
    print(x)
    return x


@App('python', dfk)
def square(x):
    if x > 5:
        return x**2
    else:
        time.sleep(5)
        return x**2


@App('python', dfk)
def increment(x):
    time.sleep(1)
    return x + 1


def test_unordered_sequence():
    r = rand().result()
    s = square(r)
    i = increment(r)
    print(s.done())
    while s.done() is not True and i.done() is not True:
        pass
    if i.done() is True:
        print(i.result())
        print(square(i.result()).result())
    elif s.done() is True:
        print(s.result())
        print(increment(s.result()).result())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", "--input", default="4",
                        action="store", dest="x", type=int)
    args = parser.parse_args()
    test_unordered_sequence()



"""Test bash apps."""
import parsl
from parsl import *

print("Parsl version: ", parsl.__version__)

import os
import time
import shutil
import argparse

# parsl.set_stream_logger()

workers = IPyParallelExecutor()
dfk = DataFlowKernel(workers)


@App('bash', dfk)
def echo_to_file(inputs=[], outputs=[], stderr='std.err', stdout='std.out'):
    pass


@App('bash', dfk)
def foo(x, y, stdout=None):
    return """echo {0} {1}
    """


def test_command_format_1():
    """Test command format for BashApps."""
    stdout = os.path.abspath('std.out')
    if os.path.exists(stdout):
        os.remove(stdout)

    app_fu = foo(1, 4, stdout=stdout)
    print("App_fu : ", app_fu)
    contents = None

    assert app_fu.result() == 0, "BashApp exited with an error code : {0}".format(
        app_fu.result())

    with open(stdout, 'r') as stdout_f:
        contents = stdout_f.read()
        print("Contents : ", contents)

    if os.path.exists('stdout_file'):
        os.remove(stdout)

    assert contents == '1 4\n', 'Output does not match expected string "1 4", Got: "{0}"'.format(
        contents)
    return True


def test_parallel_for(n=10):
    """Test a simple parallel for loop."""
    outdir = os.path.abspath('outputs')
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    else:
        shutil.rmtree(outdir)
        os.makedirs(outdir)

    d = {}

    start = time.time()
    for i in range(0, n):
        d[i], _ = echo_to_file(inputs=['Hello World {0}'.format(i)],
                               outputs=['{0}/out.{1}.txt'.format(outdir, i)],
                               stdout='{0}/std.{1}.out'.format(outdir, i),
                               stderr='{0}/std.{1}.err'.format(outdir, i),
                               )
        # time.sleep(0.01)

    assert len(
        d.keys()) == n, "Only {0}/{1} keys in dict".format(len(d.keys()), n)

    [d[i].result() for i in d]
    print("Duration : {0}s".format(time.time() - start))
    stdout_file_count = len(
        [item for item in os.listdir(outdir) if item.endswith('.out')])
    assert stdout_file_count == n, "Only {0}/{1} files in '{1}' ".format(len(os.listdir('outputs/')),
                                                                         n, outdir)
    print("[TEST STATUS] test_parallel_for [SUCCESS]")
    return d


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    # x = test_parallel_for(int(args.count))
    y = test_command_format_1()
    # raise_error(0)



# A point in a workflow process where one or more activities can
# be repeated

from parsl import *
import random
import argparse

workers = ThreadPoolExecutor(max_workers=10)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    x = random.randint(2, 10)
    print(x)
    print("random")
    return square(x).result()


@App('python', dfk)
def square(x):
    y = x**2
    if y > 25:
        return increment(y).result()
    else:
        return cubed(x).result()


@App('python', dfk)
def cubed(x):
    y = x**3
    return square(y).result()


@App('python', dfk)
def increment(x):
    return x + 1


def test_arbitrary(x=2):
    numbers = []
    for i in range(x):
        numbers.append(rand())
    for i in range(x):
        print(numbers[i].result())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", "--num", default="2",
                        action="store", dest="x", type=int)
    args = parser.parse_args()
    test_arbitrary(args.x)




from parsl import *
import argparse

workers = ThreadPoolExecutor(max_workers=6)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def get_num(first, second):
    return first + second


def test_fibonacci(num=5):
    x1 = 0
    x2 = 1
    counter = 0
    results = []
    results.append(0)
    results.append(1)
    while counter < num - 2:
        counter += 1
        results.append(get_num(x1, x2))
        temp = x2
        x2 = get_num(x1, x2)
        x1 = temp
    for i in range(len(results)):
        if isinstance(results[i], int):
            print(results[i])
        else:
            print(results[i].result())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--num", default="5",
                        action="store", dest="a", type=int)
    args = parser.parse_args()
    test_fibonacci(args.a)



import uuid
import time
import hashlib
import os
import getpass
import json
import logging
import socket

from parsl.dataflow.states import States

logger = logging.getLogger(__name__)


class UsageTracker (object):
    """Anonymized Usage Tracking for Parsl.

    Client for this is here : https://github.com/Parsl/parsl_tracking
    This issue captures the discussion that went into functionality
    implemented here : https://github.com/Parsl/parsl/issues/34

    """

    def __init__(self, dfk, ip='52.3.111.203', port=50077,
                 domain_name='tracking.parsl-project.org'):
        """Initialize usage tracking unless the user has opted-out.

        Tracks usage stats by inspecting the internal state of the dfk.

        Args:
             - dfk (DFK object) : Data Flow Kernel object

        KWargs:
             - ip (string) : IP address
             - port (int) : Port number, Default:50077
             - domain_name (string) : Domain name, will override IP
                  Default: tracking.parsl-project.org
        """
        if domain_name:
            try:
                self.UDP_IP = socket.gethostbyname(domain_name)
            except Exception:
                logging.debug("Could not lookup domain_name, defaulting to 52.3.111.203")
                self.UDP_IP = ip
        else:
            self.UDP_IP = ip
        self.UDP_PORT = port
        self.dfk = dfk
        self.config = self.dfk.config
        self.uuid = str(uuid.uuid4())
        self.test_mode, self.tracking_enabled = self.check_tracking_enabled()
        logger.debug("Tracking status: {}".format(self.tracking_enabled))
        logger.debug("Testing mode   : {}".format(self.test_mode))
        self.initialized = False  # Once first message is sent this will be True

    def check_tracking_enabled(self):
        """By default tracking is enabled.

        If Test mode is set via env variable PARSL_TESTING, a test flag is set

        Tracking is disabled if :
            1. config["globals"]["usageTracking"] is set to False (Bool)
            2. Environment variable PARSL_TRACKING is set to false (case insensitive)

        """
        track = True   # By default we track usage
        test = False  # By default we are not in testing mode

        testvar = str(os.environ.get("PARSL_TESTING", 'None')).lower()
        if testvar == 'true':
            test = True

        if self.config and self.config["globals"]["usageTracking"] is False:
            track = False

        envvar = str(os.environ.get("PARSL_TRACKING", True)).lower()
        if envvar == "false":
            track = False

        return test, track

    def construct_start_message(self):
        """Collect preliminary run info at the start of the DFK.

        Returns :
              - Message dict dumped as json string, ready for UDP
        """
        uname = getpass.getuser().encode('latin1')
        hashed_username = hashlib.sha256(uname).hexdigest()[0:10]
        hname = socket.gethostname().encode('latin1')
        hashed_hostname = hashlib.sha256(hname).hexdigest()[0:10]
        message = {'uuid': self.uuid,
                   'uname': hashed_username,
                   'hname': hashed_hostname,
                   'test': self.test_mode,
                   'start': time.time()}

        return json.dumps(message)

    def construct_end_message(self):
        """Collect the final run information at the time of DFK cleanup.

        Returns:
             - Message dict dumped as json string, ready for UDP
        """
        app_count = self.dfk.task_count

        site_count = 0
        if self.dfk._executors_managed:
            site_count = len(self.dfk.config['sites'])

        failed_states = (States.failed, States.dep_fail)
        app_fails = len([t for t in self.dfk.tasks if
                         self.dfk.tasks[t]['status'] in failed_states])

        message = {'uuid': self.uuid,
                   'end': time.time(),
                   't_apps': app_count,
                   'sites': site_count,
                   'c_time': None,
                   'failed': app_fails,
                   'test': self.test_mode,
                   }

        return json.dumps(message)

    def send_UDP_message(self, message):
        """Send UDP message."""
        if self.tracking_enabled:
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # UDP
                x = sock.sendto(bytes(message, "utf-8"), (self.UDP_IP, self.UDP_PORT))
                sock.close()
            except OSError:
                logger.debug("Unable to reach the network to send usage data")
                x = 0
        else:
            x = -1

        return x

    def send_message(self):
        """Send message over UDP.

        If tracking is disables, the bytes_sent will always be set to -1

        Returns:
            (bytes_sent, time_taken)
        """
        start = time.time()
        message = None
        if not self.initialized:
            message = self.construct_start_message()
            self.initialized = True
        else:
            message = self.construct_end_message()

        x = self.send_UDP_message(message)
        end = time.time()

        return x, end - start


if __name__ == '__main__':

    from parsl import *

    workers = ThreadPoolExecutor(max_workers=4)
    dfk = DataFlowKernel(executors=[workers])

    # ut = UsageTracker(dfk, ip='52.3.111.203')
    ut = UsageTracker(dfk, domain_name='tracking.parsl-project.org')

    for i in range(0, 2):
        x = ut.send_message()
        print(x)



import parsl
import time
from parsl import *
import argparse

config = {
    "sites": [
        {"site": "Local_Threads",
         "auth": {"channel": None},
         "execution": {
             "executor": "threads",
             "provider": None,
             "maxThreads": 2,
         }
        }],
    "globals": {"lazyErrors": True,
                "memoize": True,
                "checkpointMode": "task_exit",
    }
}
dfk = DataFlowKernel(config=config)


@App('python', dfk, cache=True)
def slow_double(x, sleep_dur=1):
    import time
    time.sleep(sleep_dur)
    return x * 2


def test_at_task_exit(n=4):
    """Test checkpointing at task_exit behavior
    """

    d = {}

    print("Launching : ", n)
    for i in range(0, n):
        d[i] = slow_double(i)
    print("Done launching")

    for i in range(0, n):
        d[i].result()

    time.sleep(1)

    print("Rundir : ", dfk.rundir)
    with open("{}/parsl.log".format(dfk.rundir), 'r') as f:
        lines = [line for line in f.readlines() if "Checkpointing.." in line]
        assert len(lines) == n, "Expected {} checkpoint events, got {}".format(n, len(lines))


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    x = test_at_task_exit(n=4)



# A point in the workflow were two or more separate branches, that are not
# being executed in parallel,  come together into one branch

from parsl import *
import argparse

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def square(x):
    return x**2


@App('python', dfk)
def increment(x):
    return x + 1


@App('python', dfk)
def cubed(x):
    return x**3


@App('python', dfk)
def join(x, y, z):
    return x + y + z


def test_join(x=2):
    print(join(square(x).result(), increment(
        x).result(), cubed(x).result()).result())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--num", default="2",
                        action="store", dest="a", type=int)
    args = parser.parse_args()
    test_join(args.a)



"""Testing bash apps
"""
import parsl
import parsl.app.errors as perror
import argparse
from parsl import *

print("Parsl version: ", parsl.__version__)

# parsl.set_stream_logger()
workers = ThreadPoolExecutor(max_workers=8)
dfk = DataFlowKernel(executors=[workers])


@App('bash', dfk)
def echo_to_streams(msg, stderr='std.err', stdout='std.out'):
    return 'echo "{0}"; echo "{0}" >&2'.format(msg)


def _test_bad_stdout():
    """Testing bad stdout file
    """
    stdout = "/x/test_bad_stdout.stdout"
    stderr = "test_bad_stdout.stderr"
    fu = echo_to_streams("Hello world", stderr=stderr, stdout=stdout)

    try:
        fu.result()
    except Exception as e:
        assert isinstance(
            e, perror.BadStdStreamFile), "Expected BadStdStreamFile, got :{0}".format(type(e))

    return


def _test_bad_stderr():
    """Testing bad stderr file
    """
    stdout = "test_bad_stdout.stdout"
    stderr = "/x/test_bad_stdout.stderr"
    fu = echo_to_streams("Hello world", stderr=stderr, stdout=stdout)

    try:
        fu.result()
    except Exception as e:
        assert isinstance(
            e, perror.BadStdStreamFile), "Expected BadStdStreamFile, got :{0}".format(type(e))

    return


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    y = test_bad_stdout()
    y = test_bad_stderr()



# An activity in a workflow process is enabled after the completion of
# another activity in the same process

from parsl import *
import random
import argparse

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    x = random.randint(1, 1000)
    return x


@App('python', dfk)
def multiply_rand(x):
    return x * random.randint(1, 10)


def test_sequence(x=5):
    flights = []
    miles = []
    for i in range(x):
        flights.append(rand())
    for j in range(len(flights)):
        miles.append(multiply_rand(flights[j].result()))
    for k in range(len(miles)):
        print(miles[k].result())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--flight", default="5",
                        action="store", dest="a", type=int)
    args = parser.parse_args()
    test_sequence(args.a)



import parsl
from parsl import *
import os
import time
import argparse
import subprocess
import sys

fn = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'test_python_checkpoint_1.py')
subprocess.check_output([sys.executable, fn])

time.sleep(0.2)
last_checkpoint = os.path.abspath(
    'runinfo/{0}/checkpoint'.format(sorted(os.listdir('runinfo/'))[-1]))
for i in range(0, 10):
    if not os.path.exists(last_checkpoint):
        print("Waiting for path")
        time.sleep(0.1)
    else:
        print("Path exists to checkpoint")
        break

config = {
    "sites": [
        {"site": "Local_Threads",
         "auth": {"channel": None},
         "execution": {
             "executor": "threads",
             "provider": None,
             "maxThreads": 2,
         }
         }],
    "globals": {"lazyErrors": True,
                "memoize": True,
    }
}

dfk = DataFlowKernel(config=config, checkpointFiles=[last_checkpoint])


@App('python', dfk, cache=True)
def slow_double(x, sleep_dur=1):
    import time
    time.sleep(sleep_dur)
    return x * 2


def test_initial_checkpoint_write(n=4):
    """2. Load the memoization table from previous checkpoint
    """

    d = {}

    start = time.time()
    print("Launching : ", n)
    for i in range(0, n):
        d[i] = slow_double(i)
    print("Done launching")

    for i in range(0, n):
        d[i].result()
    print("Done sleeping")

    delta = time.time() - start
    assert delta < 1, "Took longer than a second, restore from checkpoint failed"


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    x = test_initial_checkpoint_write(n=4)



from parsl import *

workers = ThreadPoolExecutor(max_workers=4)
data_flow_kernel = DataFlowKernel(executors=[workers])


@App('bash', data_flow_kernel)
def echo(message, outputs=[]):
    return 'echo {0} &> {outputs[0]}'


@App('python', data_flow_kernel)
def cat(inputs=[]):
    with open(inputs[0].filepath) as f:
        return f.readlines()


def test_slides():
    """Testing code snippet from slides """

    hello = echo("Hello World!", outputs=['hello1.txt'])
    message = cat(inputs=[hello.outputs[0]])

    # Waits. This need not be in the slides.
    print(hello.result())
    print(message.result())



"""Testing bash apps
"""
import parsl
from parsl import *

print("Parsl version: ", parsl.__version__)


# parsl.set_stream_logger()
workers = ThreadPoolExecutor(max_workers=8)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def generate(limit):
    from random import randint
    """Generate a random integer and return it"""
    return randint(1, limit)


def test_parallel_for(N=5):
    """Test parallel workflows from docs on Composing workflows
    """
    rand_nums = []
    for i in range(1, 5):
        rand_nums.append(generate(i))

    # wait for all apps to finish and collect the results
    outputs = [i.result() for i in rand_nums]
    return outputs


if __name__ == "__main__":

    test_parallel_for()



import os

USERNAME = os.environ['BEAGLE_USERNAME']

"""
================== Block
| ++++++++++++++ | Node
| |            | |
| |    Task    | |             . . .
| |            | |
| ++++++++++++++ |
==================
"""

singleNode = {
    "sites": [
        {"site": "Remote_IPP",
         "auth": {
             "channel": "ssh",
             "hostname": "login4.beagle.ci.uchicago.edu",
             "username": USERNAME,
             "scriptDir": "/lustre/beagle2/{}/parsl_scripts".format(USERNAME)
         },
         "execution": {
             "executor": "ipp",
             "provider": "torque",
             "block": {  # Definition of a block
                 "nodes": 1,            # of nodes in that block
                 "launcher": 'aprun',
                 "taskBlocks": 1,       # total tasks in a block
                 "initBlocks": 1,
                 "maxBlocks": 1,
                 "options": {
                     "partition": "debug",
                     "overrides": """export PATH=/lustre/beagle2/yadunand/bin/anaconda3/bin:$PATH
which python3;
source activate parsl_env_3.6.3;
"""
                 }
             }
         }
         }
    ],
    "globals": {"lazyErrors": True}
}


"""
                      Block {Min:0, init:1, Max:1}
========================================================================
| ++++++++++++++ || ++++++++++++++ || ++++++++++++++ || ++++++++++++++ |
| |    Node    | || |    Node    | || |    Node    | || |    Node    | |
| |            | || |            | || |            | || |            | |
| | Task  Task | || | Task  Task | || | Task  Task | || | Task  Task | |
| |            | || |            | || |            | || |            | |
| ++++++++++++++ || ++++++++++++++ || ++++++++++++++ || ++++++++++++++ |
========================================================================

"""
multiNodeSrun = {
    "sites": [
        {"site": "Local_IPP",
         "auth": {
             "channel": "ssh",
             "hostname": "beagle.nersc.gov",
             "username": USERNAME,
             "scriptDir": "/global/homes/y/{}/parsl_scripts".format(USERNAME),
         },
         "execution": {
             "executor": "ipp",
             "provider": "slurm",  # LIKELY SHOULD BE BOUND TO SITE
             "block": {  # Definition of a block
                 "launcher": "srun",
                 "nodes": 4,            # of nodes in that block
                 "taskBlocks": 8,       # total tasks in a block
                 "walltime": "00:10:00",
                 "initBlocks": 1,
                 "maxBlocks": 1,
                 "options": {
                     "partition": "debug",
                     "overrides": """#SBATCH --constraint=haswell
module load python/3.5-anaconda ; source activate parsl_env_3.5"""
                 }
             }
         }
         }
    ],
    "globals": {"lazyErrors": True}
}


"""
                      Block {Min:0, init:1, Max:1}
========================================================================
| ++++++++++++++ || ++++++++++++++ || ++++++++++++++ || ++++++++++++++ |
| |    Node    | || |    Node    | || |    Node    | || |    Node    | |
| | +--------------------------+ | || | +--------------------------+ | |
| | |         MPI Task         | | || | |         MPI Task         | | |
| | +--------------------------+ | || | +--------------------------+ | |
| ++++++++++++++ || ++++++++++++++ || ++++++++++++++ || ++++++++++++++ |
========================================================================

"""
multiNodeMPI = {
    "sites": [
        {"site": "Remote_IPP_MultiNode",
         "auth": {
             "channel": "ssh",
             "hostname": "beagle.nersc.gov",
             "username": USERNAME,
             "scriptDir": "/global/homes/y/{}/parsl_scripts".format(USERNAME)
         },
         "execution": {
             "executor": "ipp",
             "provider": "slurm",
             "block": {  # Definition of a block
                 "launcher": "srun",
                 "nodes": 4,            # of nodes in that block
                 "taskBlocks": 2,       # total tasks in a block
                 "walltime": "00:10:00",
                 "initBlocks": 1,
                 "maxBlocks": 1,
                 "options": {
                     "partition": "debug",
                     "overrides": """#SBATCH --constraint=haswell
module load python/3.5-anaconda ; source activate parsl_env_3.5"""
                 }
             }
         }
         }
    ],
    "globals": {"lazyErrors": True}
}



localThreads = {
    "sites": [
        {"site": "Local_Threads",
         "auth": {"channel": None},
         "execution": {
             "executor": "threads",
             "provider": None,
             "maxThreads": 4
         }
        }],
    "globals": {"lazyErrors": True}
}

localIPP = {
    "sites": [
        {"site": "Local_IPP",
         "auth": {"channel": None},
         "execution": {
             "executor": "ipp",
             "provider": "local",
             "block": {
                 "initBlocks": 4,  # Start with 4 workers
             }
         }
        }],
    "globals": {"lazyErrors": True}
}



# A point in the workflow process where a single thread of control splits into multiple
# threads of control which can be executed in parallel, this allows calls to be executed either
# simultaneously or in any order

from parsl import *
import random
import argparse

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    return random.randint(1, 10)


@App('python', dfk)
def square(x):
    return x**2


@App('python', dfk)
def cubed(x):
    return x**3


def test_parallel_split(x=4):
    for i in range(x):
        num = rand().result()
        y = square(num).result()
        z = cubed(num).result()
        print(y)
        print(z)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--input", default="4",
                        action="store", dest="a", type=int)
    args = parser.parse_args()
    test_parallel_split(args.a)



#!/usr/bin/env python3.5
from nose.tools import assert_raises

import parsl
from parsl import *

parsl.set_stream_logger()

from parsl.executors.swift_t import *


def foo(x, y):
    return x * y


def slow_foo(x, y):
    import time
    time.sleep(x)
    return x * y


def bad_foo(x, y):
    time.sleep(x)
    return x * y


def test_simple():
    print("Start")
    tex = TurbineExecutor()
    x = tex.submit(foo, 5, 10)
    print("Got: ", x)
    print("X result: ", x.result())
    assert x.result() == 50, "X != 50"
    print("done")


def test_slow():
    futs = {}
    tex = TurbineExecutor()
    for i in range(0, 3):
        futs[i] = tex.submit(slow_foo, 1, 2)

    total = sum([futs[i].result(timeout=10) for i in futs])
    assert total == 6, "expected 6, got {}".format(total)


def test_except():
    def get_bad_result():
        tex = TurbineExecutor()
        x = tex.submit(bad_foo, 5, 10)

        return x.result()

    assert_raises(NameError, get_bad_result)


if __name__ == "__main__":

    # test_simple()
    # test_slow()
    test_except()

    print("Done")



import os
from parsl import *


def test_2():
    """From docs. Attempt loading checkpoint from previous run
    """

    from parsl.configs.local import localThreads as config
    last_runid = sorted(os.listdir('runinfo/'))[-1]
    last_checkpoint = os.path.abspath('runinfo/{0}/checkpoint'.format(last_runid))

    dfk = DataFlowKernel(config=config,
                         checkpointFiles=[last_checkpoint])

    # Test addition
    dfk.cleanup()



# A point in a workflow process where an instance of an activity is removed completely

from parsl import *
import random
import argparse

workers = ThreadPoolExecutor(max_workers=10)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    x = random.randint(1, 10)
    print(x)
    return x


@App('python', dfk)
def square(x):
    z = x**2
    return z


@App('python', dfk)
def increment(x):
    return x + 1


@App('python', dfk)
def cubed(x):
    return x ** 3


@App('python', dfk)
def sum_elements(x=[], y=[], z=[]):
    total = 0
    for i in range(len(x)):
        total += x[i].result()
    for j in range(len(y)):
        total += y[j].result()
    for k in range(len(z)):
        total += z[k].result()
    return total


def test_withdraw(x=3):
    cubes = []
    squares = []
    increments = []
    for i in range(x):
        r = rand().result()
        cubes.append(cubed(r))
        squares.append(square(r))
        increments.append(increment(r))
    r = random.randint(1, 30)
    print(r)
    if r > 20:
        del cubes[:]
    if r < 20 and r > 10:
        del squares[:]
    if r < 10:
        del increments[:]
    print(sum_elements(cubes, squares, increments).result())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", "--input", default="3",
                        action="store", dest="x", type=int)
    args = parser.parse_args()
    test_withdraw(args.x)



"""Define the File Type.

The primary purpose of the File object is to track the protocol to be used
to transfer the file as well as to give the appropriate filepath depending
on where(client-side, remote-side, intermediary-side) the File.filepath is
being called from
"""

import os
import logging
from urllib.parse import urlparse
from parsl.data_provider.data_manager import DataManager


logger = logging.getLogger(__name__)


class File(object):
    """The Parsl File Class.

    This is planned to be a very simple class that simply
    captures various attributes of a file, and relies on client-side and worker-side
    systems to enable to appropriate transfer of files.
    """

    def __init__(self, url, to=None, site=None, dman=None, cache=False, caching_dir=".", staging='direct'):
        """Construct a File object from a url string.

        Args:
             - url (string) : url string of the file e.g.
               'input.txt'
               'file:///scratch/proj101/input.txt'
               'globus://go#ep1/~/data/input.txt'
               'globus://ddb59aef-6d04-11e5-ba46-22000b92c6ec/home/johndoe/data/input.txt'
               'ftp://ftp.sra.ebi.ac.uk/vol1/ERA156/ERA156312/fastq/LC_C14_cRNA_sequence_R1.txt.gz'
             - dman (DataManager) : data manager
             - to (string) : path the remote file will be staged to
             - site (string) : site the remote file will be staged to
        """
        self.url = url
        parsed_url = urlparse(self.url)
        self.scheme = parsed_url.scheme if parsed_url.scheme else 'file'
        self.netloc = parsed_url.netloc
        self.path = parsed_url.path
        self.filename = os.path.basename(self.path)
        self.to = to if to else '.'
        self.dman = dman if dman else DataManager.get_data_manager()
        self.data_future = {}
        if self.scheme != 'file':
            self.dman.add_file(self)

        self.cache = cache
        self.caching_dir = caching_dir
        self.staging = staging

    def __str__(self):
        return self.filepath

    def __repr__(self):
        return self.__str__()

    @property
    def filepath(self):
        """Returns the resolved filepath on the side where it is called from.

        File.filepath returns the appropriate filepath when called from within
        an app running remotely as well as regular python on the client side.

        Args:
            - self
        Returns:
             - filepath (string)
        """
        if self.scheme == 'globus':
            if hasattr(self, 'local_path'):
                return self.local_path

        if 'exec_site' not in globals() or self.staging == 'direct':
            # Assume local and direct
            return self.path
        else:
            # Return self.path for now
            return self.path

    def stage_in(self, site=None):
        ''' The stage_in call transports the file from the site of origin
        to the local site
        '''
        return self.dman.stage_in(self, site)

    def stage_out(self):
        ''' The stage_out call transports the file from local filesystem
        to the origin site
        '''
        return self.dman.stage_out(self)

    def set_data_future(self, df, site=None):
        self.data_future[site] = df

    def get_data_future(self, site):
        return self.data_future.get(site)


if __name__ == '__main__':

    x = File('./files.py')



"""Execution Provider Factory.

Centralize creation of execution providers and executors.

"""

import copy
import logging

# Executors
from parsl.executors.ipp import IPyParallelExecutor
from parsl.executors.swift_t import TurbineExecutor
from parsl.executors.threads import ThreadPoolExecutor
from parsl.execution_provider.errors import *

# Controller
from parsl.executors.ipp_controller import Controller

# Execution Providers and channels
from libsubmit import *

logger = logging.getLogger(__name__)


class ExecProviderFactory (object):

    def __init__(self):
        """Constructor for the execution provider factory.

        Args:
             None
        """
        self.executors = {'ipp': IPyParallelExecutor,
                          'swift_t': TurbineExecutor,
                          'threads': ThreadPoolExecutor,
                          None: lambda *args, **kwargs: None}

        self.execution_providers = {'local': Local,
                                    # Cloud Systems
                                    'aws': EC2Provider,
                                    'googleCloud': GoogleCloud,
                                    'azure': AzureProvider,
                                    # Cluster/HPC systems
                                    'slurm': Slurm,
                                    'cobalt': Cobalt,
                                    'condor': Condor,
                                    'torque': Torque,
                                    'gridEngine': GridEngine,
                                    None: lambda *args, **kwargs: None}

        self.channels = {'ssh': SshChannel,
                         'ssh-il': SshILChannel,
                         'local': LocalChannel,
                         None: lambda *args, **kwargs: None}

    def validate_config(self, config):
        """Validate configuration.

        There is no logic implemented here yet.
        This might be a good first task for a new dev.

        Args:
             - config (dict) : Config data structure
        Returns:
             - Bool: validity of config
        """
        return True

    def make(self, rundir, config):
        """Construct the appropriate provider, executors and channels and link them together."""
        self.rundir = rundir
        sites = {}

        for site in config.get("sites"):

            logger.debug("Constructing site : %s ", site.get('site', 'Unnamed_site'))
            channel_name = site["auth"]["channel"]

            if channel_name in self.channels:
                channel_opts = site["auth"].copy()
                if "channel" in channel_opts:
                    del channel_opts["channel"]
                channel = self.channels[channel_name](**channel_opts)

            else:
                logger.error("Site:{0} requests an invalid channel:{0}".format(site["site"],
                                                                               channel_name))
                raise BadConfig(site["site"],
                                "invalid channel:{0} requested".format(channel_name))

            logger.debug("Created channel : {0}".format(channel))

            provider_name = site["execution"]["provider"]
            if provider_name in self.execution_providers:
                provider = self.execution_providers[provider_name](site,
                                                                   channel=channel)

            else:
                logger.error("Site:{0} requests an invalid provider:{0}".format(site["site"],
                                                                                provider_name))
                raise BadConfig(site["site"],
                                "invalid provider:{0} requested".format(provider_name))

            logger.debug("Created execution_provider : {0}".format(provider))

            executor_name = site["execution"]["executor"]

            if executor_name in self.executors:

                controller = None

                if executor_name == 'ipp' and config.get("controller", None):

                    logger.debug("Starting controller")
                    # A controller needs to be started per run
                    site["controller"] = copy.copy(config["controller"])

                    site["controller"]['ipythonDir'] = self.rundir
                    site["controller"]['profile'] = config["controller"].get('profile', site["site"])

                    controller = Controller(**site["controller"])
                    logger.debug("Controller engine file : %s", controller.engine_file)
                    logger.debug("Controller client file : %s", controller.client_file)

                executor = self.executors[executor_name](execution_provider=provider,
                                                         controller=controller,
                                                         config=site)

            else:
                logger.error("Site:{0} requests an invalid executor:{0}".format(site["site"],
                                                                                executor_name))
                raise BadConfig(site["site"],
                                "invalid executor:{0} requested".format(executor_name))

            logger.debug("Created executor : {0}".format(executor))

            sites[site["site"]] = executor

        return sites



"""Testing bash apps
"""
import parsl
from parsl import *

print("Parsl version: ", parsl.__version__)


# parsl.set_stream_logger()
workers = ThreadPoolExecutor(max_workers=8)
dfk = DataFlowKernel(executors=[workers])


@App('bash', dfk)
def generate(outputs=[]):
    return "echo $(( RANDOM % (10 - 5 + 1 ) + 5 )) &> {outputs[0]}"


@App('bash', dfk)
def concat(inputs=[], outputs=[], stdout="stdout.txt", stderr='stderr.txt'):
    return "cat {0} >> {1}".format(" ".join(map(lambda x: x.filepath, inputs)), outputs[0])


@App('python', dfk)
def total(inputs=[]):
    total = 0
    with open(inputs[0].filepath, 'r') as f:
        for l in f:
            total += int(l)
    return total


def test_parallel_dataflow():
    """Test parallel dataflow from docs on Composing workflows
    """

    # create 5 files with random numbers
    output_files = []
    for i in range(5):
        output_files.append(generate(outputs=['random-%s.txt' % i]))

    # concatenate the files into a single file
    cc = concat(inputs=[i.outputs[0]
                        for i in output_files], outputs=["all.txt"])

    # calculate the average of the random numbers
    totals = total(inputs=[cc.outputs[0]])
    print(totals.result())


if __name__ == "__main__":

    test_parallel_dataflow()



# In this pattern, multiple instances of an activity can be enabled, with the
# number of instances of a given activity known at design time

from parsl import *
import random
import argparse

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    x = random.randint(1, 10)
    print(x)
    return x


@App('python', dfk)
def square(x):
    return x**2


@App('python', dfk)
def increment_one(x):
    return x + 1


@App('python', dfk)
def sum_elements(x, y, z):
    return x + y + z


def test_multi_instances():
    x = increment_one(square(rand().result()).result()).result()
    y = increment_one(square(rand().result()).result()).result()
    z = increment_one(square(rand().result()).result()).result()
    total = sum_elements(x, y, z).result()
    print(total)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", "--int", default="5",
                        action="store", dest="x", type=int)
    args = parser.parse_args()
    test_multi_instances()



import getpass
import os

HOME = os.environ["HOME"]
USERNAME = getpass.getuser()

"""
================== Block
| ++++++++++++++ | Node
| |            | |
| |    Task    | |             . . .
| |            | |
| ++++++++++++++ |
==================
"""

singleNodeLocal = {
    "sites": [
        {"site": "Local_CC-IN2P3",
         "auth": {
             "channel": "local",
             "username": USERNAME,
             "scriptDir": "{}/parsl_scripts".format(HOME)
         },
         "execution": {
             "executor": "ipp",
             "provider": "gridEngine",
             "block": {  # Definition of a block
                 "nodes": 1,            # of nodes in that block
                 "taskBlocks": 1,       # total tasks in a block
                 "initBlocks": 1,
                 "maxBlocks": 1,
                 "options": {
                     "partition": "debug",
                     "overrides": """export PATH=/pbs/throng/lsst/software/anaconda/anaconda3-5.0.1/bin:$PATH;
source activate parsl_env_3.5"""
                 }
             }
         }
         }
    ],
    "globals": {"lazyErrors": True}
}

"""
                      Block {Min:0, init:1, Max:1}
========================================================================
| ++++++++++++++ || ++++++++++++++ || ++++++++++++++ || ++++++++++++++ |
| |    Node    | || |    Node    | || |    Node    | || |    Node    | |
| |            | || |            | || |            | || |            | |
| | Task  Task | || | Task  Task | || | Task  Task | || | Task  Task | |
| |            | || |            | || |            | || |            | |
| ++++++++++++++ || ++++++++++++++ || ++++++++++++++ || ++++++++++++++ |
========================================================================

"""
multiNodeLocal = {
    "sites": [
        {"site": "Local_CC-IN2P3",
         "auth": {
             "channel": "local",
             "username": USERNAME,
             "scriptDir": "{}/parsl_scripts".format(HOME)
         },
         "execution": {
             "executor": "ipp",
             "provider": "gridEngine",
             "block": {  # Definition of a block
                 "taskBlocks": 2,       # total tasks in a block
                 "initBlocks": 1,
                 "maxBlocks": 4,
                 "options": {
                     "partition": "debug",
                     "overrides": """export PATH=/pbs/throng/lsst/software/anaconda/anaconda3-5.0.1/bin:$PATH;
source activate parsl_env_3.5"""
                 }
             }
         }
         }
    ],
    "globals": {"lazyErrors": True}
}



"""Testing bash apps
"""
import parsl
from parsl import *

print("Parsl version: ", parsl.__version__)

import os
import time
import shutil
import argparse

# parsl.set_stream_logger()
workers = ThreadPoolExecutor(max_workers=8)

# workers = ProcessPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('bash', dfk)
def echo_to_file(inputs=[], outputs=[], stderr='std.err', stdout='std.out'):
    cmd_line = 'echo {inputs[0]} > {outputs[0]}'
    return cmd_line


def test_parallel_for(n=10):
    """Testing a bash outputs in a parallel for
    """
    outdir = 'outputs'
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    else:
        shutil.rmtree(outdir)
        os.makedirs(outdir)

    d = {}

    start = time.time()
    for i in range(0, n):
        d[i] = echo_to_file(inputs=['Hello World {0}'.format(i)],
                            outputs=['{0}/out.{1}.txt'.format(outdir, i)],
                            stdout='{0}/std.{1}.out'.format(outdir, i),
                            stderr='{0}/std.{1}.err'.format(outdir, i),
                            )
        # time.sleep(0.01)

    assert len(
        d.keys()) == n, "Only {0}/{1} keys in dict".format(len(d.keys()), n)

    [d[i].result() for i in d]
    print([d[i].outputs for i in d])
    print("Duration : {0}s".format(time.time() - start))
    stdout_file_count = len(
        [item for item in os.listdir(outdir) if item.endswith('.out')])
    assert stdout_file_count == n, "Only {0}/{1} files in '{1}' ".format(len(os.listdir('outputs/')),
                                                                         n, outdir)
    print("[TEST STATUS] test_parallel_for [SUCCESS]")
    return d


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    x = test_parallel_for(int(args.count))
    # raise_error(0)



from parsl import *
import os

os.environ["COOLEY_USERNAME"] = "yadunand"
from cooley import singleNode as config
dfk = DataFlowKernel(config=config)


@App("python", dfk)
def test():
    import platform
    return "Hello from {0}".format(platform.uname())


results = {}
for i in range(0, 5):
    results[i] = test()

print("Waiting ....")
print(results[0].result())



import logging
import time
import math

logger = logging.getLogger(__name__)


class Strategy (object):
    """FlowControl Strategy.

    As a workflow dag is processed by Parsl, new tasks are added and completed
    asynchronously. Parsl interfaces executors with execution providers to construct
    scalable execution sites to handle the variable work-load generated by the
    workflow. This component is responsible for periodically checking outstanding
    tasks and available compute capacity and trigger scaling events to match
    workflow needs.

    Here's a diagram of a site. A site consists of blocks, which are usually
    created by single requests to a Local Resource Manager (LRM) such as slurm,
    condor, torque, or even AWS API. The blocks could contain several task blocks
    which are separate instances on workers.


    .. code:: python

                |<--minBlocks     |<-initBlocks              maxBlocks-->|
                +--------------------------------------------------------+
                |  +--------Block--------+       +--------Block--------+ |
         Site = |  | TaskBlock TaskBlock | ...   | TaskBlock TaskBlock | |
                |  +---------------------+       +---------------------+ |
                +--------------------------------------------------------+

    The general shape and bounds of a site are user specified through:

       1. minBlocks: Minimum # of blocks to maintain per site
       2. initBlocks: # of blocks to provision at initialization of workflow
       3. maxBlocks: Maximum # of blocks that can be active at a site from one workflow.


    .. code:: python

          slots = current_capacity * taskBlocks

          active_tasks = pending_tasks + running_tasks

          Parallelism = slots / tasks
                      = [0, 1] (i.e,  0 <= p <= 1)

    For eg:

    When p = 0,
         => compute with the least resources possible.
         infinite tasks are stacked per slot.

         .. code:: python

               blocks =  minBlocks           { if active_tasks = 0
                         max(minBlocks, 1)   {  else

    When p = 1,
         => compute with the most resources.
         one task is stacked per slot.

         .. code:: python

               blocks = min ( maxBlocks,
                        ceil( active_tasks / slots ) )


    When p = 1/2,
         => We stack upto 2 tasks per slot before we overflow
         and request a new block


    let's say min:init:max = 0:0:4 and taskBlocks=2

    In the diagram, X <- task

    at 2 tasks :

    .. code:: python

        +---Block---|
        |           |
        | X      X  |
        |slot   slot|
        +-----------+

    at 5 tasks, we overflow as the capacity of a single block is fully used.

    .. code:: python

        +---Block---|       +---Block---|
        | X      X  | ----> |           |
        | X      X  |       | X         |
        |slot   slot|       |slot   slot|
        +-----------+       +-----------+

    """

    def __init__(self, dfk):
        """Initialize strategy."""
        self.dfk = dfk
        self.config = dfk.config
        self.sites = {}
        self.max_idletime = 60 * 2  # 2 minutes

        for site in self.dfk.config["sites"]:
            self.sites[site['site']] = {'idle_since': None,
                                        'config': site}

        self.strategies = {None: self._strategy_noop,
                           'simple': self._strategy_simple}

        strtgy_name = self.config['globals'].get('strategy', None)
        self.strategize = self.strategies.get(strtgy_name,
                                              self._strategy_noop)

        logger.debug("Scaling strategy: {0}".format(strtgy_name))

    def _strategy_noop(self, tasks, *args, kind=None, **kwargs):
        """Do nothing.

        Args:
            - tasks (task_ids): Not used here.

        KWargs:
            - kind (Not used)
        """

    def _strategy_simple(self, tasks, *args, kind=None, **kwargs):
        """Peek at the DFK and the sites specified.

        We assume here that tasks are not held in a runnable
        state, and that all tasks from an app would be sent to
        a single specific site, i.e tasks cannot be specified
        to go to one of more sites.

        Args:
            - tasks (task_ids): Not used here.

        KWargs:
            - kind (Not used)
        """
        # Add logic to check sites
        # for task in tasks :
        #    if self.dfk.tasks[task]:

        for sitename in self.dfk.executors:

            exc = self.dfk.executors[sitename]
            site_config = self.sites[sitename]['config']

            if not exc.scaling_enabled:
                logger.debug("Site:{0} Status:STATIC".format(sitename))
                continue

            # Tasks that are either pending completion
            active_tasks = exc.executor.outstanding

            # Get the status of the taskBlocks
            status = exc.status()

            # Get the shape and bounds for the site
            minBlocks = site_config["execution"]["block"]["minBlocks"]
            maxBlocks = site_config["execution"]["block"]["maxBlocks"]
            initBlocks = site_config["execution"]["block"]["initBlocks"]
            taskBlocks = site_config["execution"]["block"]["taskBlocks"]
            parallelism = site_config["execution"]["block"]["parallelism"]

            active_blocks = sum([1 for x in status if x in ('RUNNING',
                                                            'SUBMITTING',
                                                            'PENDING')])
            active_slots = active_blocks * taskBlocks

            logger.debug("Min:{} initBlocks:{} Max:{}".format(minBlocks,
                                                              initBlocks,
                                                              maxBlocks))
            logger.debug("Tasks:{} Slots:{} Parallelism:{}".format(len(active_tasks),
                                                                   active_slots,
                                                                   parallelism))

            # Case 1
            # No tasks.
            if len(active_tasks) == 0:
                # Case 1a
                # Fewer blocks that minBlocks
                if active_blocks <= minBlocks:
                    # Ignore
                    # logger.debug("Strategy: Case.1a")
                    pass

                # Case 1b
                # More blocks than minBlocks. Scale down
                else:
                    # We want to make sure that max_idletime is reached
                    # before killing off resources
                    if not self.sites[sitename]['idle_since']:
                        logger.debug("Strategy: Scale_in, tasks=0 starting kill timer")
                        self.sites[sitename]['idle_since'] = time.time()

                    idle_since = self.sites[sitename]['idle_since']
                    if (time.time() - idle_since) > self.max_idletime:
                        # We have resources idle for the max duration,
                        # we have to scale_in now.
                        logger.debug("Strategy: Scale_in, tasks=0")
                        exc.scale_in(active_blocks - minBlocks)

                    else:
                        pass
                        # logger.debug("Strategy: Case.1b. Waiting for timer : {0}".format(idle_since))

            # Case 2
            # More tasks than the available slots.
            elif (float(active_slots) / len(active_tasks)) < parallelism:
                # Case 2a
                # We have the max blocks possible
                if active_blocks >= maxBlocks:
                    # Ignore since we already have the max nodes
                    # logger.debug("Strategy: Case.2a")
                    pass

                # Case 2b
                else:
                    # logger.debug("Strategy: Case.2b")
                    excess = math.ceil((len(active_tasks) * parallelism) - active_slots)
                    excess_blocks = math.ceil(float(excess) / taskBlocks)
                    logger.debug("Requesting : {}".format(excess_blocks))
                    exc.scale_out(excess_blocks)

            elif active_slots == 0 and len(active_tasks) > 0:
                # Case 4
                # Check if slots are being lost quickly ?
                logger.debug("Requesting single slot")
                exc.scale_out(1)
            # Case 3
            # tasks ~ slots
            else:
                # logger.debug("Strategy: Case 3")
                pass


if __name__ == '__main__':

    pass



from parsl import *

import os
import time
import shutil
import argparse

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('bash', dfk)
def multi_line(inputs=[], outputs=[], stderr='std.err', stdout='std.out'):
    cmd_line = """echo {inputs[0]} &> {outputs[0]}
    echo {inputs[1]} &> {outputs[1]}
    echo {inputs[2]} &> {outputs[2]}
    echo "Testing STDOUT"
    echo "Testing STDERR" 1>&2
    """
    return cmd_line


def run_test():

    outdir = 'outputs'
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    else:
        shutil.rmtree(outdir)
        os.makedirs(outdir)

    f = multi_line(inputs=["Hello", "This is", "Cat!"],
                   outputs=['{0}/hello.txt'.format(outdir),
                            '{0}/this.txt'.format(outdir),
                            '{0}/cat.txt'.format(outdir)])
    print(f.result())

    time.sleep(0.1)
    assert 'hello.txt' in os.listdir(outdir), "hello.txt is missing"
    assert 'this.txt' in os.listdir(outdir), "this.txt is missing"
    assert 'cat.txt' in os.listdir(outdir), "cat.txt is missing"
    with open('std.out', 'r') as o:
        out = o.read()
        assert out != "Testing STDOUT", "Stdout is bad"
    with open('std.err', 'r') as o:
        err = o.read()
        assert err != "Testing STDERR", "Stderr is bad"

    os.remove('std.err')
    os.remove('std.out')
    return True


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    # if args.debug:
    #    parsl.set_stream_logger()

    run_test()



from parsl import *
import os
import parsl
import getpass

parsl.set_stream_logger()

os.environ["COOLEY_USERNAME"] = getpass.getuser()

from cooley import singleNodeLocal as config
dfk = DataFlowKernel(config=config)


@App("bash", dfk)
def freesurfer(stdout=None, stderr=None):
    return """singularity exec ~madduri/freesurfer.simg recon-all
    """


if __name__ == "__main__":

    N = 4
    results = {}
    for i in range(0, N):
        results[i] = freesurfer(stdout="freesurfer.{}.out".format(i),
                                stderr="freesurfer.{}.err".format(i))

    for i in range(0, N):
        results[i].result()

    print("Waiting ....")
    try:
        print(results[0].result())
    except Exception as e:
        print("Caught an exception, but this is not a problem")
        pass
    print("STDOUT from 0th run :")
    print(open(results[0].stdout, 'r').read())
    print(open(results[0].stderr, 'r').read())



import logging

from parsl.app.futures import DataFuture
from parsl.app.app import AppBase

logger = logging.getLogger(__name__)


class PythonApp(AppBase):
    """Extends AppBase to cover the Python App."""

    def __init__(self, func, executor, walltime=60, cache=False,
                 sites='all', fn_hash=None):
        """Initialize the super.

        This bit is the same for both bash & python apps.
        """
        super().__init__(func, executor, walltime=walltime, sites=sites, exec_type="python")
        self.fn_hash = fn_hash
        self.cache = cache

    def __call__(self, *args, **kwargs):
        """This is where the call to a python app is handled.

        Args:
             - Arbitrary
        Kwargs:
             - Arbitrary

        Returns:
             If outputs=[...] was a kwarg then:
                   App_fut, [Data_Futures...]
             else:
                   App_fut

        """
        app_fut = self.executor.submit(self.func, *args,
                                       parsl_sites=self.sites,
                                       fn_hash=self.fn_hash,
                                       cache=self.cache,
                                       **kwargs)

        # logger.debug("App[{}] assigned Task[{}]".format(self.func.__name__,
        #                                                 app_fut.tid))
        out_futs = [DataFuture(app_fut, o, parent=app_fut, tid=app_fut.tid)
                    for o in kwargs.get('outputs', [])]
        app_fut._outputs = out_futs

        return app_fut



import parsl
from parsl import *
import os

os.environ['SWAN_USERNAME'] = 'p01953'
from swan import singleNode as config
parsl.set_stream_logger()
dfk = DataFlowKernel(config=config)


@App("python", dfk)
def python_app_slow(duration):
    import platform
    import time
    time.sleep(duration)
    return "Hello from {0}".format(platform.uname())


def test_python_remote(count=10):
    """ Run with no delay
    """
    fus = []
    for i in range(0, count):
        fu = python_app_slow(0)
        fus.extend([fu])

    for fu in fus:
        print(fu.result())


def test_python_remote_slow(count=10):

    fus = []
    for i in range(0, count):
        fu = python_app_slow(count / 10)
        fus.extend([fu])

    for fu in fus:
        print(fu.result())


if __name__ == "__main__":

    test_python_remote()
    test_python_remote_slow()



import sys
import threading
import logging
import time

from parsl.dataflow.strategy import Strategy

logger = logging.getLogger(__name__)


class FlowNoControl(object):
    """FlowNoControl implements similar interfaces as FlowControl.

    Null handlers are used so as to mimic the FlowControl class.

    """

    def __init__(self, dfk, config, *args, threshold=2, interval=2):
        """Initialize the flowcontrol object. This does nothing.

        Args:
             - dfk (DataFlowKernel) : DFK object to track parsl progress
             - config (dict) : Config dict structure

        KWargs:
             - threshold (int) : Tasks after which the callback is triggered
             - interval (int) : seconds after which timer expires
        """
        pass

    def notify(self, event_id):
        """This notifiy fn does nothing."""
        pass

    def close(self):
        """This close fn does nothing."""
        pass


class FlowControl(object):
    """Implements threshold-interval based flow control.

    The overall goal is to trap the flow of apps from the
    workflow, measure it and redirect it the appropriate executors for
    processing.

    This is based on the following logic :

    .. code-block:: none

        BEGIN (INTERVAL, THRESHOLD, callback) :
            start = current_time()

            while (current_time()-start < INTERVAL) :
                 count = get_events_since(start)
                 if count >= THRESHOLD :
                     break

            callback()

    This logic ensures that the callbacks are activated with a maximum delay
    of INTERVAL for systems with infrequent events as well as systems which would
    generate large bursts of events.

    Once a callback is triggered, the callback generally runs a strategy
    method on the sites available as well asqeuque


    TODO : When the debug logs are enabled this module emits duplicate messages.
    This issue needs more debugging. What I've learnt so far is that the duplicate
    messages are present only when the timer thread is started, so this could be
    from a duplicate logger being added by the thread.

    """

    def __init__(self, dfk, config, *args, threshold=20, interval=5):
        """Initialize the flowcontrol object.

        We start the timer thread here

        Args:
             - dfk (DataFlowKernel) : DFK object to track parsl progress
             - config (dict) : Config dict structure

        KWargs:
             - threshold (int) : Tasks after which the callback is triggered
             - interval (int) : seconds after which timer expires
        """
        self.dfk = dfk
        self.threshold = threshold
        self.interval = interval
        self.cb_args = args
        self.strategy = Strategy(dfk)
        self.callback = self.strategy.strategize
        self._handle = None
        self._event_count = 0
        self._event_buffer = []
        self._wake_up_time = time.time() + 1
        self._kill_event = threading.Event()
        self._thread = threading.Thread(target=self._wake_up_timer, args=(self._kill_event,))
        self._thread.daemon = True
        self._thread.start()

    def _wake_up_timer(self, kill_event):
        """Internal. This is the function that the thread will execute.
        waits on an event so that the thread can make a quick exit when close() is called

        Args:
            - kill_event (threading.Event) : Event to wait on
        """

        # Sleep till time to wake up
        while True:
            prev = self._wake_up_time

            # Waiting for the event returns True only when the event
            # is set, usually by the parent thread
            time_to_die = kill_event.wait(float(max(prev - time.time(), 0)))

            if time_to_die:
                return

            if prev == self._wake_up_time:
                self.make_callback(kind='timer')
            else:
                print("Sleeping a bit more")

    def notify(self, event_id):
        """Let the FlowControl system know that there is an event."""
        self._event_buffer.extend([event_id])
        self._event_count += 1
        if self._event_count >= self.threshold:
            logger.debug("Eventcount >= threshold")
            self.make_callback(kind="event")

    def make_callback(self, kind=None):
        """Makes the callback and resets the timer.

        KWargs:
               - kind (str): Default=None, used to pass information on what
                 triggered the callback
        """
        self._wake_up_time = time.time() + self.interval
        self.callback(tasks=self._event_buffer, kind=kind)
        self._event_buffer = []

    def close(self):
        """Merge the threads and terminate."""
        self._kill_event.set()
        self._thread.join()


class Timer(object):
    """This timer is a simplified version of the FlowControl timer.
    This timer does not employ notify events.

    This is based on the following logic :

    .. code-block:: none


        BEGIN (INTERVAL, THRESHOLD, callback) :
            start = current_time()

            while (current_time()-start < INTERVAL) :
                 wait()
                 break

            callback()

    """

    def __init__(self, callback, *args, interval=5):
        """Initialize the flowcontrol object
        We start the timer thread here

        Args:
             - dfk (DataFlowKernel) : DFK object to track parsl progress
             - config (dict) : Config dict structure

        KWargs:
             - threshold (int) : Tasks after which the callback is triggered
             - interval (int) : seconds after which timer expires
        """

        self.interval = interval
        self.cb_args = args
        self.callback = callback
        self._wake_up_time = time.time() + 1

        self._kill_event = threading.Event()
        self._thread = threading.Thread(target=self._wake_up_timer, args=(self._kill_event,))
        self._thread.daemon = True
        self._thread.start()

    def _wake_up_timer(self, kill_event):
        """Internal. This is the function that the thread will execute.
        waits on an event so that the thread can make a quick exit when close() is called

        Args:
            - kill_event (threading.Event) : Event to wait on
        """

        # Sleep till time to wake up
        while True:
            prev = self._wake_up_time

            # Waiting for the event returns True only when the event
            # is set, usually by the parent thread
            time_to_die = kill_event.wait(float(max(prev - time.time(), 0)))

            if time_to_die:
                return

            if prev == self._wake_up_time:
                self.make_callback(kind='timer')
            else:
                print("Sleeping a bit more")

    def make_callback(self, kind=None):
        """Makes the callback and resets the timer.
        """
        self._wake_up_time = time.time() + self.interval
        self.callback(*self.cb_args)

    def close(self):
        """Merge the threads and terminate.
        """
        self._kill_event.set()
        self._thread.join()


if __name__ == "__main__":

    def foo():
        print("Callback made at :", time.time())

    timer = Timer(foo)

    time.sleep(60)
    timer.close()
    exit(0)

    print("This is broken")

    def cback(*args):
        print("*" * 40)
        print("Callback at {0} with args : {1}".format(time.time(), args))
        print("*" * 40)

    fc = FlowControl(cback)

    print("Testing")
    print("Press E(Enter) to create and event, X(Enter) to exit")
    while True:
        x = sys.stdin.read(1)
        if x.lower() == 'e':
            print("Event")
            fc.notify()
        elif x.lower() == 'x':
            print("Exiting ...")
            break
        else:
            print("Continuing.. got[%s]", x)



from setuptools import setup, find_packages

with open('parsl/version.py') as f:
    exec(f.read())

with open('requirements.txt') as f:
    install_requires = f.readlines()

# tests_require = parse_requirements('test-requirements.txt')

setup(
    name='parsl',
    version=VERSION,
    description='Simple data dependent workflows in Python',
    long_description='Simple and easy parallel workflows system for Python',
    url='https://github.com/Parsl/parsl',
    author='Yadu Nand Babuji',
    author_email='yadu@uchicago.edu',
    license='Apache 2.0',
    download_url='https://github.com/Parsl/parsl/archive/{}.tar.gz'.format(VERSION),
    package_data={'': ['LICENSE']},
    packages=find_packages(),
    install_requires=install_requires,
    classifiers=[
        # Maturity
        'Development Status :: 3 - Alpha',
        # Intended audience
        'Intended Audience :: Developers',
        # Licence, must match with licence above
        'License :: OSI Approved :: Apache Software License',
        # Python versions supported
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords=['Workflows', 'Scientific computing'],
)



"""Testing python outputs
"""
from parsl import *
import os
import shutil
import argparse

# parsl.set_stream_logger()
workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def double(x, outputs=[]):
    with open(outputs[0], 'w') as f:
        f.write(x * 5)
    return x * 5


def launch_apps(n, dirpath):

    outdir = dirpath
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    else:
        shutil.rmtree(outdir)
        os.makedirs(outdir)

    all_futs = {}
    for i in range(n):

        fus, _ = double(i, outputs=['{0}/{1}.txt'.format(dirpath, i)])
        print(fus.outputs)
        all_futs[fus] = fus

    stdout_file_count = len(
        [item for item in os.listdir(outdir) if item.endswith('.txt')])
    assert stdout_file_count == n, "Only {0}/{1} files in '{1}' ".format(len(os.listdir('outputs/')),
                                                                         n, outdir)
    print("[TEST STATUS] test_parallel_for [SUCCESS]")


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    args = parser.parse_args()

    x = launch_apps(10, "outputs")
    # raise_error(0)



import os
from parsl import *


def test_1():
    """Checkpointing example from docs.
    """
    from parsl.configs.local import localThreads as config
    dfk = DataFlowKernel(config=config)

    @App('python', dfk, cache=True)
    def slow_double(x, sleep_dur=1):
        import time
        time.sleep(sleep_dur)
        return x * 2

    N = 5   # Number of calls to slow_double
    d = []  # List to store the futures
    for i in range(0, N):
        d.append(slow_double(i))

    # Wait for the results
    [i.result() for i in d]

    cpt_dir = dfk.checkpoint()
    print(cpt_dir)  # Prints the checkpoint dir

    # Testing component
    assert os.path.exists(cpt_dir), "Checkpoint dir does not exist"
    dfk.cleanup()



# #In this pattern, multiple instances of an activity can be enabled, but the
# number of instances of a given activity is not known at any point during design
# time or runtime.

from parsl import *
import random
import argparse

workers = ThreadPoolExecutor(max_workers=4)
dfk = DataFlowKernel(executors=[workers])


@App('python', dfk)
def rand():
    x = random.randint(1, 10)
    return x


@App('python', dfk)
def increment_one(x):
    return x + 1


@App('python', dfk)
def sum_elements(x=[]):
    total = 0
    for i in range(len(x)):
        total += x[i]
    return total


def test_multi_instances():
    numbers = []
    while sum_elements(numbers).result() < 20:
        numbers.append(increment_one(rand().result()).result())
    print(sum_elements(numbers).result())


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", "--int", default="5",
                        action="store", dest="x", type=int)
    args = parser.parse_args()
    test_multi_instances()
